//#define EBATWO
//#define ADDON2
#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
using System.Reflection;
using MVR.FileManagementSecure;
#endregion

namespace Cotyounoyume{
	public class EBATCore : MVRScript
	{
		public float version = 1.81f;

		#region common parameters
		public Atom atom;
		public SimpleJSON.JSONClass settingsRoot = new SimpleJSON.JSONClass();
		public SimpleJSON.JSONClass settings = new SimpleJSON.JSONClass();
		public SimpleJSON.JSONClass sceneSettings = new SimpleJSON.JSONClass();
		public string myPath;
		public bool maintenanceAlert = false;
		public float lastTouchingTime = 0f;
		public string gender;
		public bool orgasmModeOn = false;
		public bool addon2ModeOn = false;
		public float perSecondOffset = 1f;
		public float lastTime = 0f;
		
		protected float openingTime;
		public float currentTime;
		public int gauge = 0;
		public float calmingDownTime;
		public float calmingDownTimeOffset = 2;
		public float calmingDownTimeMinusOffset;
		public int calmingDownPlus;
		public bool isHot = false;
		public int orgasmCount = 0;
		public int ecstasyScale = 5;
		
		public int curStage = -1;
		public int preStage = -1;
		protected string defaultSettingsPath = "Custom/Scripts/cotyounoyume/ExpressionBlushingAndTears/Config/defaultSettings.json";
		protected string currentSettingsPath = "Custom/Scripts/ExpressionBlushingAndTears_currentSettings.json";
		protected string myClassName = "EBATCore";

		public string pluginLabel;
		public bool disableLog = true;
		public int[] stageThreshold;
		public bool isBreathActive = false;
		public bool isDuringOrgasm = false;
		public bool isDuringAfterglow = false;
		public bool isEmbarassed = false;		
		public bool isFirstTerm = false;
		public bool isFirstWave = false;
		public bool hasMultipleOrgasms = false;
		public bool isFirstSquirt = true;
		public string myName = "";
		
		public bool rNippleItemActive = false;
		public bool lNippleItemActive = false;
		public bool vaginaItemActive = false;
		public bool labiaItemActive = false;
		public bool collisionFailure = false;
		
		
		public Blushing blushing;
		public Tears tears;
		public EyesControl eyesControl;
		public Belly belly;
		public Etc etc;
		public BodyMorph bodyMorph;
		public UnityAssets unityAssets;
		public WorldUI worldUi;
		public Sound sound;
		public CustomUI customUi;
		public Expression2 expression2;
		public Guess guess;
		public HeadController headController;
		public CustomUnityAssetItem customUnityAssetItem;
		public CollisionFailureDetector collisionFailureDetector;
		public Hint hint;
		
		public bool menuOnGlobal = true;
		public List<Atom> generatedAtoms = new List<Atom>();
		public List<string> generatedAtomNames = new List<string>();
		public Dictionary<string, Atom> generatedAtomsDictionary = new Dictionary<string, Atom>();
		public List<GameObject> generatedGameObjects = new List<GameObject>();
		public List<string> generatedGameObjectNames = new List<string>();
		public Dictionary<string, GameObject> generatedGameObjectsDictionary = new Dictionary<string, GameObject>();
		public AddonManager1 addonManager1;
		
		public string launchMode = "normal";
		public Atom trunk;
		public Atom trunkLow;
		public Atom floor;
		public float homingDelay = 1.0f;
		
		public SphereCollider labiaCollider = null;
		public bool collider2xOnGlobal = false;
		
		public bool NeedForNotification = false;
#if	(EBATWO)
		public Orgasm orgasm;
#endif
#if	(ADDON2)
		public AddonManager2 addonManager2;
#endif
		#endregion
				
		public override void Init(){
			try{
                if (containingAtom.type != "Person")
                {
                    SuperController.LogError("This plug-in have to be added to person type atom.");
                    return;
                }
				atom = containingAtom;
#if	(EBATWO)
				orgasmModeOn = true;
#endif
#if	(ADDON2)
				addon2ModeOn = true;
#endif
				
				JSONStorable geometry = atom.GetStorableByID("geometry");
				DAZCharacterSelector character = geometry as DAZCharacterSelector;
				gender = character.gender.ToString();
				
				myName = atom.name;
				logger("atom name: " + myName);
				myPath = Utility.GetPluginPath(this as MVRScript, atom);
				logger("myPath: " + myPath);
				
				if(Utility.HasThisPlugin(atom, "ExpressionBlushingAndTears/PAUSED")){
					launchMode = "paused";
				}
				
				blushing    = new Blushing(this);
				tears       = new Tears(this);
				bodyMorph   = new BodyMorph(this);
				eyesControl = new EyesControl(this);
				belly       = new Belly(this);
				etc         = new Etc(this);
				sound       = new Sound(this);
				unityAssets = new UnityAssets(this);
				addonManager1 = new AddonManager1(this);
				DoAddonManager2Function("Construct");
				worldUi     = new WorldUI(this, 0.001f);
				customUi    = new CustomUI(this);
				expression2 = new Expression2(this);
				guess       = new Guess(this);
				headController = new HeadController(this);
				customUnityAssetItem = new CustomUnityAssetItem(this);
				collisionFailureDetector = new CollisionFailureDetector(this);
				hint        = new Hint(this);
#if	(EBATWO)	
				orgasm      = new Orgasm(this);
#endif
				InitCore();
			}catch (Exception e){
				loggerE(e);
			}
		}

		public void syncParameter(){
			isDuringOrgasm = false;
			isDuringAfterglow = false;
#if	(EBATWO)
			isDuringOrgasm = orgasm.isDuringOrgasm;
			isDuringAfterglow = orgasm.isDuringAfterglow;
#endif
		}
		
		public void InitCore(){			
			if(configFileExist() == false){
				loadSettings("default");
				saveConfig();
			}else{
				loadSettings("current");				
			}
			applySettings();
			customUi.InitExpandState(false);
			customUi.SetupUIVariables();
			customUi.setupUI();
			guess.SetTrunk();
			guess.SetTrunkLow();
		}
		
		public void addSettingsNew(){
			SimpleJSON.JSONClass defaultSettingsRoot = new SimpleJSON.JSONClass();
			defaultSettingsRoot = SuperController.singleton.LoadJSON(defaultSettingsPath) as JSONClass;
			foreach(var key in defaultSettingsRoot.Keys){
				JSONNode jsn1 = defaultSettingsRoot[key];
				JSONNode jsn2 = settingsRoot[key];
				SyncJSON(key, jsn1, jsn2);
			}
//			debug("settingsRoot:" + settingsRoot.ToString());
		}
		
		public void SyncJSON(string keyOrig, JSONNode jsn1, JSONNode jsn2){
//			debug("keyOrig:" + keyOrig);
//			debug("jsn1:" + jsn1.ToString());
//			debug("jsn2:" + jsn2.ToString());
			if(jsn1 is JSONClass){
//				debug("JSONClass route.");
				JSONClass jsc1 = jsn1 as JSONClass;
				foreach(var key in jsc1.Keys){
					JSONNode newjsn1 = jsc1[key];
					JSONNode newjsn2 = jsn2[key];
					if((keyOrig == "General" && key == "Label") ||
					   (keyOrig == "General" && key == "Version")){
						jsn2[key] = newjsn1;
					}else if(newjsn2 == null){
//						debug("unmatch: jsn1 HasKey, but jsn2 doesn't have key: " + key);
						jsn2.Add(key, newjsn1);
					}else{
//						debug("match: jsn1 and jsn2 HasKey: " + key);
						SyncJSON(key, jsn1[key], jsn2[key]);
					}
				}
			}else if(jsn1 is JSONArray){
//				debug("JSONArray route.");
				JSONArray jsa1 = jsn1 as JSONArray;
				for(int key = 0; key < jsn1.Count; key++){
					string keyStr = key.ToString();
					JSONNode newjsn1 = jsa1[key];
					JSONNode newjsn2 = jsn2[key];
					if(newjsn2 == null){
//						debug("unmatch: jsn1 HasKey, but jsn2 doesn't have key: " + keyStr);
						jsn2.Add(keyStr, newjsn1);
					}else{
//						debug("match: jsn1 and jsn2 HasKey: " + keyStr);
						SyncJSON(keyStr, jsn1[keyStr], jsn2[keyStr]);
					}
				}				
			}else{
				if(jsn1.Count > 0){
					debug("ERROR: unknown JSONNode: key:" + keyOrig + ", value:" + jsn1.ToString());				
				}
			}
		}
		
		public void loadSettings(string type){
			try{
				defaultSettingsPath = myPath + "/" + "Config/defaultSettings.json";
				string path = (type == "default") ? defaultSettingsPath : currentSettingsPath;
				settingsRoot = SuperController.singleton.LoadJSON(path) as JSONClass;
				settings = (gender == "Female") ? settingsRoot["Female"].AsObject : settingsRoot["Male"].AsObject;

				if(version != settingsRoot["Common"]["General"]["Version"].AsFloat){
					addSettingsNew();
					NeedForNotification = true;
				}
				
				#region general settings from json
				pluginLabel = settingsRoot["Common"]["General"]["Label"];
				disableLog = settingsRoot["Common"]["General"]["Debug Message Off"].AsBool;

				menuOnGlobal = settings["General"]["Display UI Menu"].AsBool;
				
				blushing.disableLog = disableLog;
				tears.disableLog = disableLog;
				bodyMorph.disableLog = disableLog;
				eyesControl.disableLog = disableLog;
				belly.disableLog = disableLog;
				headController.disableLog = disableLog;
				etc.disableLog = disableLog;
				sound.disableLog = disableLog;
				
				stageThreshold = new int[6];
				stageThreshold[0] = 0;
				ecstasyScale = settings["General"]["Ecstasy Scale"].AsInt;
				collider2xOnGlobal = settings["General"]["2x Collider Radius"].AsBool;

				if(collider2xOnGlobal && CheckLabiaTriggerNormal()){
					ToggleDoublingColliderSize();
				}else if(!collider2xOnGlobal && !CheckLabiaTriggerNormal()){
					ToggleDoublingColliderSize();
				}
				
				for(int i = 0; i < 6; i++){
					stageThreshold[i] = settings["StageThreshold"][("Threshold" + i)]["Value"].AsInt * ecstasyScale;				
				}
				CheckLabiaTriggerNormal();
				#endregion
				
				#region 
				expression2.loadSettings();
				blushing.loadSettings();
				tears.loadSettings();
				bodyMorph.loadSettings();
				eyesControl.loadSettings();
				belly.loadSettings();
				etc.loadSettings();
				sound.loadSettings();
				unityAssets.loadSettings();
				customUi.loadSettings();
				worldUi.loadSettings();
				guess.loadSettings();
				headController.loadSettings();

				loadSettingsAlt();
				
				#endregion
			}catch (Exception e){
				loggerE(e);
			}
		}

		public void resetThreshold(){
			for(int i = 0; i < 6; i++){
				stageThreshold[i] = settings["StageThreshold"][("Threshold" + i)]["Value"].AsInt * ecstasyScale;				
			}
		}
		
		public void loadSettingsAlt(){
#if (EBATWO)
			orgasm.loadSettings();
			orgasm.applySettings();
			orgasm.maintenance();
#endif
			addonManager1.loadSettings();
			addonManager1.applySettings();			
			DoAddonManager2Function("LoadSettings");
			DoAddonManager2Function("ApplySettings");
		}

		public void Start(){
			try{
				#region at once: base
				worldUi.makeWorldUI();
				unityAssets.makeUAPee();
				
				openingTime = Time.time;
				logger("openingTime = " + openingTime);
				logger("myPath:" + myPath);
				setGauge(0);
				isBreathActive = false;
				#endregion
				
				maintenanceSequence();
				belly.storePositionAndRotation();
				headController.storePositionAndRotation();
				tears.unequipTearsAll();
				bodyMorph.stopBodyMorph();
				expression2.stopExpression();
			}catch (Exception e){
				loggerE(e);
			}			
		}

		public bool configFileExist(){
			try{
				if(FileManagerSecure.FileExists(currentSettingsPath)){
					return true;
				}else{
					return false;
				}
			}catch (Exception e){
				loggerE(e);
				return false;
			}
		}
		
		
		public void applySettings(){
			try{
				pluginLabelJSON.val = settingsRoot["Common"]["General"]["Label"];

				#region
				blushing.applySettings();
				tears.applySettings();
				bodyMorph.applySettings();
				eyesControl.applySettings();
				sound.applySettings();
				belly.applySettings();
				headController.applySettings();
				etc.applySettings();
				unityAssets.applySettings();
				worldUi.applySettings();
				
				#endregion
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void versionCheck(){
			if(NeedForNotification){
				hint.VersionCheckHint();
				NeedForNotification = false;
				saveConfig();
			}
				
			pluginLabel = settingsRoot["Common"]["General"]["Label"];			
		}
		
		
		public void maintenanceSequence(){
			try{
				#region initialize trigger
				resetTriggers();
				RegisterActions();
				#endregion
				
				#region maintenance
				blushing.maintenance();
				tears.maintenance();
				bodyMorph.maintenance();
				eyesControl.maintenance();
				belly.maintenance();
				headController.maintenance();
				etc.maintenance();
				#endregion
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void embarrassALittle(){
			if(getGauge() < stageThreshold[2]){
				setGauge(stageThreshold[2]);
			}
			OnErogenousZoneTouch("Nipple");			
		}
		
		public void DoAddonManager2Function(string func){
#if (ADDON2)
			if(func == "Construct"){
				addonManager2 = new AddonManager2(this);
			}else if(func == "LoadSettings"){
				addonManager2.loadSettings();
			}else if(func == "ApplySettings"){
				addonManager2.applySettings();				
			}else if(func == "Manage"){
				addonManager2.manage(currentTime);				
			}else if(func == "OnDestroy"){
				addonManager2.OnDestroy();				
			}else if(func == "OnOrgasmStart"){
				addonManager2.OnOrgasmStart();				
			}else if(func == "OnOrgasmStart"){
				addonManager2.OnOrgasmStart();				
			}else{
				debug("In EBATCore: Unknown command: " + func);
			}
#endif	
		}
		
		public void saveConfig(){
			try{
				SuperController.singleton.SaveJSON(settingsRoot, currentSettingsPath);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void setGauge(int value){
			if(gauge > stageThreshold[5] * 1.2){
				value = (int)(stageThreshold[5] * 1.2);
			}
			gauge = value;
			if(gauge > 0){
				calmingDownTime = currentTime + calmingDownTimeOffset + calmingDownTimeMinusOffset;
			}else{
				calmingDownTime = currentTime + 3600 * 24;
			}
			if(getStage() >= 2){
				isHot = true;
			}else{
				isHot = false;
			}
			isStageChanged();
			if(launchMode == "normal"){
				worldUi.uiGaugeSlider.slider.value = gauge;
			}
		}
		
		public int getGauge(){
			return gauge;
		}
		
		public void incrementGauge(int value = 1){
			setGauge(gauge + value);
		}
		
		public void decrementGauge(){
			if(isDuringOrgasm){
				return;
			}
			int fragment = (int)((currentTime - lastTouchingTime)/10);
			fragment = (fragment < 0) ? 0 : fragment;
			setGauge(gauge - (1 + fragment) * ecstasyScale - calmingDownPlus);
		}
		
		#region trigger functions
		public void RegisterActions(){
			try{
//				RegisterActionPair("LipTrigger", "OnLipTouch", "Lip");
//				RegisterStartActionPair("MouthTrigger", "OnMouthTouch", "Mouth");				
//				RegisterStartActionPair("ThroatTrigger", "OnThroatTouch", "Throat");				
				RegisterStartActionPair("lNippleTrigger", "OnNippleTouch", "Nipple");
				RegisterStartActionPair("rNippleTrigger", "OnNippleTouch", "Nipple");
				RegisterStartActionPair("LabiaTrigger", "OnLabiaTouch", "Labia");
				RegisterStartActionPair("VaginaTrigger", "OnVaginaTouch", "Vagina");
				RegisterStartActionPair("DeepVaginaTrigger", "OnVaginaDeepTouch", "VaginaDeep");
				RegisterStartActionPair("DeeperVaginaTrigger", "OnVaginaDeeperTouch", "VaginaDeeper");
				
//				RegisterEndActionPair("MouthTrigger", "OnMouthTouchEnd", "Mouth");
//				RegisterEndActionPair("ThroatTrigger", "OnThroatTouchEnd", "Throat");				
				RegisterEndActionPair("lNippleTrigger", "OnNippleTouchEnd", "Nipple");
				RegisterEndActionPair("rNippleTrigger", "OnNippleTouchEnd", "Nipple");
				RegisterEndActionPair("LabiaTrigger", "OnLabiaTouchEnd", "Labia");
				RegisterEndActionPair("VaginaTrigger", "OnVaginaTouchEnd", "Vagina");
				RegisterEndActionPair("DeepVaginaTrigger", "OnVaginaDeepTouchEnd", "VaginaDeep");
				RegisterEndActionPair("DeeperVaginaTrigger", "OnVaginaDeeperTouchEnd", "VaginaDeeper");
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void RegisterEndActionPair(string collisionTrigger, string action, string position){
			RegisterStartActionPair(collisionTrigger, action, position, "end");
		}
		
		public void RegisterStartActionPair(string collisionTrigger, string action, string position, string type = "start"){
			try{
				if(type == "start"){
					JSONStorableAction tmpAction = new JSONStorableAction(action, () =>{
						OnErogenousZoneTouch(position);
					});
					RegisterAction(tmpAction);
				}else{
					JSONStorableAction tmpAction = new JSONStorableAction(action, () =>{
						OnErogenousZoneTouchEnd(position);
					});
					RegisterAction(tmpAction);
				}
				createTrigger(collisionTrigger, action, type);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void resetTriggers() {
			try{
				var table = new Dictionary<string, int>{
//					{"OnLipTouch", 1},
					{"OnMouthTouch", 1},
					{"OnThroatTouch", 1},
					{"OnNippleTouch", 1},
					{"OnLabiaTouch", 1},
					{"OnVaginaTouch", 1},
					{"OnVaginaDeepTouch", 1},
					{"OnVaginaDeeperTouch", 1}
				};
				var tableEnd = new Dictionary<string, int>{
//					{"OnLipTouch", 1},
					{"OnMouthTouchEnd", 1},
					{"OnThroatTouchEnd", 1},
					{"OnNippleTouchEnd", 1},
					{"OnLabiaTouchEnd", 1},
					{"OnVaginaTouchEnd", 1},
					{"OnVaginaDeepTouchEnd", 1},
					{"OnVaginaDeeperTouchEnd", 1}
				};
//				string[] cTriggerNames = {"LipTrigger", "MouthTrigger", "ThroatTrigger","rNippleTrigger", "lNippleTrigger", "LabiaTrigger", "VaginaTrigger", "DeepVaginaTrigger", "DeeperVaginaTrigger"};
				string[] cTriggerNames = {"MouthTrigger", "ThroatTrigger","rNippleTrigger", "lNippleTrigger", "LabiaTrigger", "VaginaTrigger", "DeepVaginaTrigger", "DeeperVaginaTrigger"};
				foreach(string cTriggerName in cTriggerNames){
					logger("cTriggerName = " + cTriggerName);
					CollisionTrigger cTrigger = atom.GetStorableByID(cTriggerName) as CollisionTrigger;
					JSONClass triggerClass = cTrigger.trigger.GetJSON();
					JSONArray triggerArray = triggerClass["startActions"].AsArray;
					for (int i=0;i< triggerArray.Count;i++) {
						logger("triggerArray[" + i + "][name]=" + triggerArray[i]["name"].Value);
						if (table.ContainsKey(triggerArray[i]["name"].Value)){
							triggerArray.Remove(i);
						}
					}
					JSONArray triggerArrayEnd = triggerClass["endActions"].AsArray;
					for (int i=0;i< triggerArrayEnd.Count;i++) {
						if (tableEnd.ContainsKey(triggerArrayEnd[i]["name"].Value)){
							triggerArrayEnd.Remove(i);
						}
					}
					cTrigger.trigger.RestoreFromJSON(triggerClass);
				}
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void triggerDeactivate(){
			string[] cTriggerNames = {"MouthTrigger", "ThroatTrigger","rNippleTrigger", "lNippleTrigger", "LabiaTrigger", "VaginaTrigger", "DeepVaginaTrigger", "DeeperVaginaTrigger"};
//			float rnd = UnityEngine.Random.Range(0.0f, 1.0f);
			foreach(string cTriggerName in cTriggerNames){
				CollisionTrigger cTrigger = atom.GetStorableByID(cTriggerName) as CollisionTrigger;
//				if(rnd < 0.25){
					cTrigger.trigger.active = false;
					cTrigger.trigger.transitionInterpValue = 0f;
//				}
			}
		}
		
		public void createTrigger(string triggerCollisionTriggerName, string scriptTriggerName, string type){
			try{
				CollisionTrigger cTrigger = atom.GetStorableByID(triggerCollisionTriggerName) as CollisionTrigger;
				cTrigger.enabled = true;
				TriggerActionDiscrete startTrigger;
				if(type == "start"){
					startTrigger=cTrigger.trigger.CreateDiscreteActionStartInternal();
				}else{
					startTrigger=cTrigger.trigger.CreateDiscreteActionEndInternal();					
				}
				startTrigger.name = scriptTriggerName;
				startTrigger.receiverAtom = atom;
				startTrigger.receiver = GetPluginStorableById(atom, myClassName);
				startTrigger.receiverTargetName = scriptTriggerName;
			}catch (Exception e){
				loggerE(e);
			}
		}
	
		// Function taken from VAMDeluxe's code
		static JSONStorable GetPluginStorableById(Atom atom, string id){
			return Utility.GetPluginStorableById(atom, id);
		}
		#endregion		
		
		#region engine
		public void OnErogenousZoneTouch(string zone){ 
			try{
				currentTime = Time.time;
				if(launchMode == "paused"){
					return;
				}
				collisionFailureDetector.registerActiveResult(zone);
				
				logger("Touch: " + zone);
				isEmbarassed = false;
				lastTouchingTime = currentTime;
				sound.playSound("moan", Time.time);
				updateAll();			

				calmingDownPlus = 0;
				calmingDownTimeMinusOffset = 0;
				int value = calcZoneMap(zone);
				incrementGauge(value);
				isStageChanged();
				
				if(isBreathActive == false){
					isBreathActive = true;
				}
				
				headController.PickAction();
				/*
				if(isDuringOrgasm && isDuringAfterglow){
					headController.PickAction();
//					headController.RandomHeadMotionForSmallWave(ratioMax:2.0f);
				}else if(isDuringOrgasm && !isDuringAfterglow){
					headController.PickAction();
//					headController.RandomHeadMotionForBigWave(ratioMax:2.0f);
				}
				*/
				orgasmFunctions("orgasmCheck", currentTime);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void OnErogenousZoneTouchEnd(string zone){ 
			
		}
		
		public int calcZoneMap(string zone){
			int stage = getStage();
			int value = 0;
			if(zone == "Mouth" || zone == "Throat"){
				if(stage <= 2){
					value = 3;
				}else if(stage <= 4){
					value = 1;
				}else{
					value = 0;
				}
			}else if(zone == "Nipple"){
				if(stage <= 2){
					value = 2;
				}else if(stage <= 4){
					value = 1;
				}else{
					value = 1;
				}
			}else if(zone == "Labia" || zone == "Vagina"){
				if(stage <= 2){
					value = 1;
				}else if(stage <= 4){
					value = 2;
				}else{
					value = 2;
				}
			}else if(zone == "VaginaDeep" || zone == "VaginaDeeper"){
				if(stage <= 2){
					value = 1;
				}else if(stage <= 4){
					value = 2;
				}else{
					value = 3;
				}
			}else if(zone == "ByScript"){
				value = 1;
			}
			return value;
		}
		
		public void OnStageChange(){
			try{
				tears.cm.unequipClothes(preStage);
				tears.cm.equipClothes(curStage);
				tears.update(currentTime);
				
				int cuaStage = getCUAStage();
				addon2Functions("SendParameterToCUA", var2:cuaStage);
				blushing.ChangeBlushEquipment();
				worldUi.UpdateStatus();
				if(addonManager1.addonTakeControl.autoStopOnGlobal){
					addonManager1.addonTakeControl.AutoStopPlugins();
				}
			}catch (Exception e){
				loggerE(e);
			}			
		}
		
		public int getCUAStage(){
			int cuaStage = 0;
			int stage = getStage();
			if(stage < 2){
				cuaStage = 0;
			}else if(stage < 3){
				cuaStage = 1;
			}else if(stage < 5){
				cuaStage = 2;
			}else{
				cuaStage = 3;
			}
			return cuaStage;
		}
		
		public void isStageChanged(){
			try{
				if(curStage != getStage()){
					preStage = curStage;
					curStage = getStage();
					bool isStageUp = (preStage < curStage) ? true : false;
					logger("stage changed: " + preStage + " => " + curStage);
					
					OnStageChange();
					if(!isHot && !isStageUp){
						OnCalmDown();
					}
				}
			}catch(Exception e){
				loggerE(e);
			}
		}
		
		public void addon2Functions(string func, string var1 = "", int var2 = 0){
#if (ADDON2)
			if(func == "SendParameterToCUA"){
				addonManager2.addonCua.SendParameterToCUA(var2);
			}
#endif	
		}
		
		public void orgasmFunctions(string func, float time){
#if (EBATWO)
			if(func == "orgasmCheck"){
				orgasm.orgasmCheck(time);
			}else if(func == "update"){
				orgasm.update(time);
			}else if(func == "orgasmControl"){
				orgasm.orgasmControl(time);
			}else if(func == "orgasmStart"){
				orgasm.orgasmStart();
			}else if(func == "orgasmFinish"){
				orgasm.orgasmFinish();
			}
#endif
		}
		
		public bool isOrgasmFunctionEnabled(){
			bool flag = false;
#if (EBATWO)
			flag = true;
#endif
			return flag;
		}
		
		public void updateAll(){
			try{
				expression2.update();
				blushing.update(currentTime);
				tears.update(currentTime);
				bodyMorph.update(currentTime);
				eyesControl.update(currentTime);
				orgasmFunctions("update", currentTime);
			}catch (Exception e){
				loggerE(e);
			}
		}
			
		public float getTime(){
			return Time.time;
		}
		
        public void Update(){
            try{
				if(Time.time > currentTime + 10){
					currentTime = Time.time;
					logger("start plug-in or load scene.");
					maintenanceSequence();
				}
				
				if(launchMode == "paused"){
					return;
				}
				
				string currentName = atom.name;
				if(myName != currentName){
					logger("old name:" + myName + ", new name:" + currentName);
					myName = currentName;
					maintenanceAlert = true;
				}
				
				if(maintenanceAlert){
					maintenanceAlert = false;
					Init();
					customUi.RemoveConfigMenu();
					Start();
					return;
				}
				
				currentTime = Time.time;
				PerSecondCheck();

				if(calmingDownTime < currentTime){
					decrementGauge();
				}
				
				orgasmFunctions("orgasmControl", currentTime);
				blushing.manage(currentTime);
				tears.manage(currentTime);
				bodyMorph.manage(currentTime);
				expression2.manage();
				eyesControl.manage(currentTime);
				belly.manage(currentTime);
				headController.manage(currentTime);
				unityAssets.manage(currentTime);
				addonManager1.manage(currentTime);
				DoAddonManager2Function("Manage");
				collisionFailureDetector.update();

#if (EBATWO)				
				if(expression2.isExpressionActive == false && blushing.isBlushingActive == false && tears.isTearsActive == false){
					if(isDuringOrgasm == false){
						setGauge(0);
						isStageChanged();
					}
				}
				
				if(isOrgasmFunctionEnabled() == false &&
					lastTouchingTime + longestDurationTime() < currentTime){
					setGauge(0);
					isStageChanged();
				}
#endif				
			}catch (Exception e){
//				loggerE(e);
			}
		}

		public void PerSecondCheck(){
			if(lastTime + perSecondOffset < currentTime){
				syncParameter();
				lastTime = currentTime;
				customUi.MemorizeUIPosition();
//				triggerDeactivate();
			}
		}
		
		public float longestDurationTime(){
			float eTime = expression2.fmm.peakTimeOffset + expression2.fmm.finishTimeOffset;
			float bTime = blushing.blushClosingTimeOffset;
			float tTime = tears.tearsClosingTimeOffset;
			float tmpTime = (eTime > bTime) ? eTime : bTime;
			float result = (tmpTime > tTime) ? tmpTime : tTime;
			return result;
		}
		
		public int getStage(){
			try{
				int stage = 0;
				int gauge = getGauge();
				if(gauge < stageThreshold[1]){
					stage = 0;
				}else if(gauge < stageThreshold[2]){
					stage = 1;
				}else if(gauge < stageThreshold[3]){
					stage = 2;
				}else if(gauge < stageThreshold[4]){
					stage = 3;
				}else if(gauge < stageThreshold[5]){
					stage = 4;
				}else{
					stage = 5;
				}
				return stage;
			}catch (Exception e){
				loggerE(e);
				return -1;
			}
		}
		
		#endregion
		
		#region etc
		public void logger(string str){
			if(disableLog == false){
				SuperController.LogMessage(str);
			}
		}
		
		public void loggerE(Exception e){
			SuperController.LogMessage("*** Exception:" + e);
		}
		
		public void debug(string str){
			SuperController.LogMessage("debug: " + str);
		}
		
		public virtual IEnumerator CreateAtom(string atomType, string prefix, Action<Atom> onAtomCreated, bool keep=false)
        {
            //  need to wait a bit before creating if we are just loading into the scene, as the previous atom may have not been fully destroyed!!
//            yield return new WaitForSeconds(1.0f);
            yield return new WaitForSeconds(0.1f);

            string atomId = GenerateAtomName(prefix, atom.uid);
            Atom tmpatom = SuperController.singleton.GetAtomByUid(atomId);
            if (tmpatom == null)
            {
                yield return SuperController.singleton.AddAtomByType(atomType, atomId);
                tmpatom = SuperController.singleton.GetAtomByUid(atomId);
            }

            if (tmpatom != null)
            {
                if (keep == false)
                {
                    generatedAtoms.Add(tmpatom);
					generatedAtomNames.Add(prefix);
					generatedAtomsDictionary.Add(atomId, tmpatom);
                }
                onAtomCreated(tmpatom);
            }
        }
		
		public virtual IEnumerator CreateGameObject(string name, PrimitiveType type, Action<GameObject> onGameObjectCreated){
			yield return new WaitForSeconds(0.1f);
			
			string objId = GenerateGameObjectName(name, atom.uid);
			GameObject tmpObj = GameObject.CreatePrimitive(type);
			if(objId != null){
				generatedGameObjects.Add(tmpObj);
				generatedGameObjectNames.Add(name);
				generatedGameObjectsDictionary.Add(objId, tmpObj);
				onGameObjectCreated(tmpObj);
			}
		}

		public string GenerateAtomName(string prefix, string suffix)
        {
            return prefix + " " + suffix;
        }

		public string GenerateGameObjectName(string prefix, string suffix)
        {
            return prefix + " go " + suffix;
        }

		public string GetGameObjectId(string objName){
			return GenerateGameObjectName(objName, atom.uid);
		}
		
		public GameObject GetGameObject(string objName){
			string objId = GetGameObjectId(objName);
			if(generatedGameObjectsDictionary.ContainsKey(objId)){
				return generatedGameObjectsDictionary[objId];
			}
			return null;
		}

		public void RemoveGameObject(string objName, bool keepName = false){
			try{
			string objId = GetGameObjectId(objName);
			
			if(generatedGameObjectsDictionary.ContainsKey(objId)){
				GameObject obj = generatedGameObjectsDictionary[objId];
				generatedGameObjects.Remove(obj);
				Destroy(obj);
				generatedGameObjectsDictionary.Remove(objId);
				if(keepName == false){
					generatedGameObjectNames.Remove(objName);
				}
			}
				
			}catch(Exception e){
				loggerE(e);
			}
		}

		public void RemoveGameObjectAll(){
			try{
				foreach(var name in generatedGameObjectNames){
					RemoveGameObject(name, true);
				}
				generatedGameObjectNames.Clear();
			}catch(Exception e){
				loggerE(e);
			}
		}
		
		public void RemoveGameObjectByName(string prefix){
			try{
				List<string> tmpList = new List<string>();
				foreach(var name in generatedGameObjectNames){
					if(name.StartsWith(prefix)){
						RemoveGameObject(name, true);
						tmpList.Add(name);
					}
				}
				foreach(var name in tmpList){
					generatedGameObjectNames.Remove(name);
				}
				tmpList.Clear();
//				generatedGameObjectNames.Clear();
			}catch(Exception e){
				loggerE(e);
			}
		}
		
		public bool CheckLabiaTriggerNormal(){
			if(labiaCollider == null){
				SphereCollider[] cols = atom.GetComponentsInChildren<SphereCollider>();
				foreach(var col in cols){
					if(col.name == "LabiaTrigger"){
						labiaCollider = col;
					}
				}
			}
			
			bool result = true;
			if(labiaCollider.radius >= 0.06f){
				result = false;
			}
			return result;
		}
		
		public void ToggleDoublingColliderSize(){
			Dictionary<string, float> collisionSize = new Dictionary<string, float>(){
				{"LabiaTrigger", 0.030f},
				{"VaginaTrigger", 0.040f},
				{"DeepVaginaTrigger", 0.035f},
				{"DeeperVaginaTrigger", 0.035f},
			};
			SphereCollider[] cols = atom.GetComponentsInChildren<SphereCollider>();
			foreach(var col in cols){
				Rigidbody rb = col.attachedRigidbody;
				if(col.name.StartsWith("Labia") || col.name.StartsWith("Vagina") || col.name.StartsWith("Deep")){
					if(collider2xOnGlobal){
						col.radius = collisionSize[col.name] * 2;
					}else{
						col.radius = collisionSize[col.name];						
					}
				}
			}			
		}
		#endregion
		
		public void OnDestroy(){
			RemoveGameObjectAll();
			finishSequence();
			worldUi.OnDestroy();
			unityAssets.OnDestroy();
			
			expression2.OnDestroy();
			tears.OnDestroy();
			blushing.OnDestroy();
			bodyMorph.OnDestroy();
			sound.OnDestroy();
			belly.OnDestroy();
			headController.OnDestroy();
			addonManager1.OnDestroy();
			collisionFailureDetector.OnDestroy();
			DoAddonManager2Function("OnDestroy");

            generatedAtoms.ForEach((atom) =>
            {
                if (atom == null)
                {
                    return;
                }
                SuperController.singleton.RemoveAtom(atom);
            });
            generatedAtoms.Clear();
			
            generatedGameObjects.ForEach((gobj) =>
            {
                if (gobj == null)
                {
                    return;
                }
				Destroy(gobj);
            });
            generatedGameObjects.Clear();			
		}
		
		public void OnCalmDown(){
//			headController.FaceForwardSlowlyForce(true);
		}
		
		public void OnOrgasmStart()
		{
			worldUi.OnOrgasmStart();
			addonManager1.OnOrgasmStart();
			DoAddonManager2Function("OnOrgasmStart");
		}
		
		public void OnOrgasmStop()
		{
			worldUi.OnOrgasmStop();
			addonManager1.OnOrgasmStop();
			DoAddonManager2Function("OnOrgasmStop");
		}
		
		public void finishSequence(){
			expression2.stopExpression();
			tears.unequipTearsAll();
			blushing.RemoveBlushing();
			bodyMorph.stopBodyMorph();			
		}
		
		public void surveyStorables(int lvl){
			List<string> ids = new List<string>();
			ids = atom.GetStorableIDs();
			for(int i = 0; i < ids.Count; i++){
				SuperController.LogMessage("id: " + ids[i]);
				List<string> names = new List<string>();
				
				if(lvl > 0){
					names = atom.GetStorableByID(ids[i]).GetBoolParamNames();
					for(int j = 0; j < names.Count; j++){
						SuperController.LogMessage("  bool: " + names[j]);
					}
				}
					
				if(lvl > 1){
					names = atom.GetStorableByID(ids[i]).GetAllParamAndActionNames();
					for(int j = 0; j < names.Count; j++){
						SuperController.LogMessage("  ParamAndAction: " + names[j]);
					}					
				}

				if(lvl > 1){
					names = atom.GetStorableByID(ids[i]).GetAllFloatAndColorParamNames();
					for(int j = 0; j < names.Count; j++){
						SuperController.LogMessage("  FloatAndColorParam: " + names[j]);
					}
				}
			}
		}

	}
	
	public class CollisionFailureDetector
	{
		public EBATCore core;
		public float failureCheckThresholdTime = 7.0f;
		public Dictionary<string, CollisionTrigger> triggerMap;
		public Dictionary<string, bool> checkOnMap;
		public Dictionary<string, bool> failureMap;
		public Dictionary<string, bool> alwaysOnMap;
		public Dictionary<string, float> lastDeactiveTimeMap;
		public Dictionary<string, float> lastCheckTimeMap;
		public string[] cTriggerNames = {"MouthTrigger", "ThroatTrigger","rNippleTrigger", "lNippleTrigger", "LabiaTrigger", "VaginaTrigger", "DeepVaginaTrigger", "DeeperVaginaTrigger"};
		public Dictionary<string, string> zoneMap = new Dictionary<string, string>(){
			{"MouthTrigger", "Mouth"}, 
			{"ThroatTrigger", "Throat"},
			{"rNippleTrigger", "Nipple"}, 
			{"lNippleTrigger", "Nipple"},
			{"LabiaTrigger", "Labia"},
			{"VaginaTrigger", "Vagina"},
			{"DeepVaginaTrigger", "VaginaDeep"},
			{"DeeperVaginaTrigger", "VaginaDeeper"},			
		};
		
		public Dictionary<string, string> zoneMapInv = new Dictionary<string, string>(){
			{"Mouth"       , "MouthTrigger"       }, 
			{"Throat"      , "ThroatTrigger"      },
			{"Nipple"      , "NippleTrigger"      }, 
			{"Labia"       , "LabiaTrigger"       },
			{"Vagina"      , "VaginaTrigger"      },
			{"VaginaDeep"  , "DeepVaginaTrigger"  },
			{"VaginaDeeper", "DeeperVaginaTrigger"},			
		};
		
		public float[] cuaStageThreshold = {
			1.2f, 1.0f, 0.8f, 0.5f,
		};
		
		public CollisionFailureDetector(EBATCore core){
			this.core = core;
			triggerMap = new Dictionary<string, CollisionTrigger>();
			checkOnMap = new Dictionary<string, bool>();
			failureMap = new Dictionary<string, bool>();
			alwaysOnMap = new Dictionary<string, bool>();
			lastCheckTimeMap = new Dictionary<string, float>();
			lastDeactiveTimeMap = new Dictionary<string, float>();
			
			Init();
		}
		
		public void Init(){
			OnDestroy();
			
			foreach(string cTriggerName in cTriggerNames){
				triggerMap.Add(cTriggerName, core.atom.GetStorableByID(cTriggerName) as CollisionTrigger);
				checkOnMap.Add(cTriggerName, false);
				failureMap.Add(cTriggerName, false);
				alwaysOnMap.Add(cTriggerName, false);
				lastCheckTimeMap.Add(cTriggerName, 0f);
				lastDeactiveTimeMap.Add(cTriggerName, 0f);
			}
		}
		
		public void OnDestroy(){
			triggerMap.Clear();
			checkOnMap.Clear();
			failureMap.Clear();
			lastCheckTimeMap.Clear();
		}

		public void subscribe(string name){
			if(!triggerMap.ContainsKey(name)){
				core.debug("ERROR: unknown trigger: " + name);
				return;
			}
			
			checkOnMap[name] = true;
			failureMap[name] = false;
			alwaysOnMap[name] = false;
			lastCheckTimeMap[name] = core.currentTime + 5f;
			lastDeactiveTimeMap[name] = core.currentTime + 5f;
		}
		
		public void unsubscribe(string name){
			if(!triggerMap.ContainsKey(name)){
				return;
			}
			
			checkOnMap[name] = false;
			failureMap[name] = false;
			alwaysOnMap[name] = false;
			lastCheckTimeMap[name] = 0f;
			lastDeactiveTimeMap[name] = 0f;
		}
		
		public void registerActiveResult(string name){
			try{
			string triggerName = zoneMapInv[name];
			if(triggerName == "NippleTrigger"){
				failureMap["r" + triggerName] = false;
				alwaysOnMap["r" + triggerName] = true;
				lastCheckTimeMap["r" + triggerName] = core.currentTime;				
				
				failureMap["l" + triggerName] = false;
				alwaysOnMap["l" + triggerName] = true;
				lastCheckTimeMap["l" + triggerName] = core.currentTime;				
			}else{
				failureMap[triggerName] = false;
				alwaysOnMap[triggerName] = true;
				lastCheckTimeMap[triggerName] = core.currentTime;				
			}
				
			}catch(Exception e){
				core.loggerE(e);
			}
		}
		
		public void update(){
//			debugPrint();
			checkCollisions();
			ringFailureCollisions();
		}
		
		public void checkCollisions(){
			foreach(var name in cTriggerNames){
				checkCollision(name);
			}
		}

		public void checkCollision(string name){
			if(!checkOnMap[name]){ return; }
			if(triggerMap[name].trigger.active){
				lastCheckTimeMap[name] = core.currentTime;
				if(core.currentTime > lastDeactiveTimeMap[name] + cuaStageThreshold[core.getCUAStage()]){
					alwaysOnMap[name] = true;
					lastDeactiveTimeMap[name] = core.currentTime;
//					core.debug("alwaysOnMap: lastCheckTime:" + lastCheckTimeMap[name]);
//					core.debug("alwaysOnMap: lastDeactiveTimeMap:" + lastDeactiveTimeMap[name]);
//					core.debug("alwaysOnMap: cuaStageThreshold:" + cuaStageThreshold[core.getCUAStage()]);
				}
			}else{
				lastDeactiveTimeMap[name] = core.currentTime;
				if(core.currentTime > lastCheckTimeMap[name] + cuaStageThreshold[core.getCUAStage()]){
					failureMap[name] = true;
					lastCheckTimeMap[name] = core.currentTime;
//					core.debug("failureMap: lastCheckTime:" + lastCheckTimeMap[name]);
//					core.debug("failureMap: lastDeactiveTimeMap:" + lastDeactiveTimeMap[name]);
//					core.debug("failureMap: cuaStageThreshold:" + cuaStageThreshold[core.getCUAStage()]);
				}
			}
		}
		
		public void debugPrint(){
			foreach(var name in cTriggerNames){
				core.debug("checkOnMap: " + name + ", " + checkOnMap[name]);
				core.debug("failureMap: " + name + ", " + failureMap[name]);
				core.debug("alwaysOnMap: " + name + ", " + alwaysOnMap[name]);
			}
		}
		
		public void ringFailureCollisions(){
//			return;
			foreach(var name in cTriggerNames){
				if(checkOnMap[name] && failureMap[name]){
					core.OnErogenousZoneTouch(zoneMap[name]);
					failureMap[name] = false;
				}
				if(checkOnMap[name] && alwaysOnMap[name]){
					triggerMap[name].trigger.active = false;
					alwaysOnMap[name] = false;
				}
			}
		}
	}
}
