#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class Blushing
	{
		#region common or UI Properties
		public Atom atom;
		public bool blushOnGlobal = true;
		public float blushClosingTimeOffset = 10f;
		public EBATCore core;
		public bool disableLog;
		public bool isBlushingActive;
		public string blushTypeDefault = "Cartoonlike-Large-Alt";
		public List<string> blushTypes;
		public Dictionary<string, string> blushMap;
		public bool isNewBlushSystemOn = true;
		public string currentBlush;
		public string previousBlush;
		
//		public Dictionary<string, DAZClothingItem> fakeMasksMap;
		public Dictionary<string, DAZClothingItem> blushMasksMap;
		public DAZCharacterSelector geometryForClothes;
		
		public Dictionary<string, string> blushOldToNewMap = new Dictionary<string, string>(){
			{"Cartoonlike-Small" , "B"},
			{"Cartoonlike-Small-Alt" , "B"},
			{"Cartoonlike-Large" , "A"},
			{"Cartoonlike-Large-Alt" , "A"},
			{"Reallike-Small" , "C"},
			{"Reallike-Small-Alt" , "C"},
			{"Reallike-Large" , "D"},
			{"Reallike-Large-Alt" , "D"},
			{"Cartoonlike-Middle" , "E"},
			{"Cartoonlike-Middle-Alt" , "E"},
		};
		#endregion
		
		#region blush parameters
		protected float blushClosingTime;
		protected JSONStorableUrl faceDecalUrl;
		protected bool blushLevelUp = false;
		Dictionary<string, UIDynamicToggle> toggles = new Dictionary<string, UIDynamicToggle>();
		#endregion

		public Blushing(EBATCore core){
			atom = core.atom;
			isBlushingActive = false;
			this.core = core;
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			geometryForClothes = geometryStorable as DAZCharacterSelector;
			previousBlush = "";
			currentBlush = "";
		}
		
		public void loadSettings(){
			try{
				blushTypeDefault = core.settings["General"]["Blush Type"];
				blushOnGlobal = core.settings["General"]["Blush On"].AsBool;
				blushClosingTimeOffset = core.settings["General"]["Blush Duration Time"].AsFloat;
				
				#region read blushing from json
				blushTypes = new List<string>();
				blushMap = new Dictionary<string, string>();
				JSONNode jsBlush = core.settings["Blush"];
				for(int i = 0; i < jsBlush.Count; i++){
					JSONNode jsPair = jsBlush[i];
					blushTypes.Add(jsPair["Name"]);
					string fullpath = core.myPath + "/Blushing/" + jsPair["File"];
					blushMap.Add(jsPair["Name"], fullpath);
				}
				
				getBlushMasks();
				core.StartCoroutine(equipAndUnequipAllMasks(blushMasksMap));
				/*
				equipAllMasks(fakeMasksMap, true);
				equipAllMasks(blushMasksMap, true);
				*/
				#endregion
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void applySettings(){
			try{
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void maintenance(){
			try{
				#region initialize blush
				JSONStorable receiver = atom.GetStorableByID("textures");
				if(receiver == null){
					core.logger("debug: receiver is null.");
					core.maintenanceAlert = true;
				}
				#endregion
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		#region blush functions
		public void ChangeBlushEquipment(){
			if(!blushOnGlobal){ return; }
			
			string type = getBlushType();
			int blushStage = getBlushStage();
			/*
			core.debug("previousBlush:" + previousBlush);
			core.debug("currentBlush:" + currentBlush);
			core.debug("type:" + type + ", blushStage:" + blushStage);
			*/
			previousBlush = currentBlush;
			if(previousBlush != ""){
				geometryForClothes.SetActiveClothingItem(blushMasksMap[previousBlush], false);
			}
			if(blushStage == 0){
				currentBlush = "";
			}else{
				currentBlush = "Type" + type + blushStage;
				geometryForClothes.SetActiveClothingItem(blushMasksMap[currentBlush], true);
			}				
		}
		
		public int getBlushStage(){
			int stage = core.getStage();
			int blushStage = 0;
			if(stage < 2){
				blushStage = 0;
			}else if(stage < 4){
				blushStage = 1;
			}else{
				if(core.hasMultipleOrgasms){
					blushStage = 3;
				}else{
					blushStage = 2;
				}
			}
			return blushStage;
		}
		
		public string getBlushType(){
			string newType = blushOldToNewMap[blushTypeDefault];
			return newType;
		}
		
		public void getBlushMasks(){
			blushMasksMap = new Dictionary<string, DAZClothingItem>();			
			List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
			for (int i = 0; i < clothes.Count; i++){
				if(clothes[i].creatorName == "cotyounoyume"){
					if(clothes[i].uid.Contains("blushMask")){
						if(clothes[i].uid.Contains("Type")){
							if(!blushMasksMap.ContainsKey(clothes[i].displayName)){
								blushMasksMap.Add(clothes[i].displayName, clothes[i]);
	//							core.debug("name:" + clothes[i].displayName);
							}
						}
					}
				}
			}
		}
		
		public void equipAllMasks(Dictionary<string, DAZClothingItem> hash, bool flag){
			if(hash == null){return;}
			foreach(KeyValuePair<string, DAZClothingItem> kvp in hash){
				geometryForClothes.SetActiveClothingItem(kvp.Value, flag);
			}
		}
		
		public IEnumerator equipAndUnequipAllMasks(Dictionary<string, DAZClothingItem> hash){
			foreach(KeyValuePair<string, DAZClothingItem> kvp in hash){
				geometryForClothes.SetActiveClothingItem(kvp.Value, true);
			}
			yield return new WaitForSeconds(0.1f);
			foreach(KeyValuePair<string, DAZClothingItem> kvp in hash){
				geometryForClothes.SetActiveClothingItem(kvp.Value, false);
			}			
		}
		
		public void switchFaceDecal(string decal){
			try{
				JSONStorable receiver = atom.GetStorableByID("textures");
				faceDecalUrl = receiver.GetUrlJSONParam("faceDecalUrl");
				core.logger("debug: switchFaceDecal: new decal = " + decal);
				faceDecalUrl.val = decal;
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public string getCurrentFaceDecal(){
			try{
				JSONStorable receiver = atom.GetStorableByID("textures");
				faceDecalUrl = receiver.GetUrlJSONParam("faceDecalUrl");
				return faceDecalUrl.val;
			}catch (Exception e){
				core.loggerE(e);
				return "";
			}
		}

		public void update(float time){
			try{
				if(isNewBlushSystemOn){
					return;
				}
				if(blushOnGlobal == false){
					return;
				}
				blushClosingTime = time + blushClosingTimeOffset;
				if(isBlushingActive == false){
					AddBlushing();
//				}else if(blushLevelUp){
//					switchFaceDecal(blushAliasHigh);
//					blushLevelUp = false;
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void manage(float time){
			try{
				if(isNewBlushSystemOn){
					return;
				}
				if(isBlushingActive && blushClosingTime < time && core.isDuringOrgasm == false && core.isHot == false){
					core.logger("blush effect stop.");
					RemoveBlushing();
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void RemoveBlushing(){
			switchFaceDecal("");
			isBlushingActive = false;
		}
		
		public void AddBlushing(){
			isBlushingActive = true;
			switchFaceDecal(blushMap[blushTypeDefault]);			
		}
		
		public void OnDestroy(){
			RemoveBlushing();
		}
		#endregion
	}
}