#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class Guess
	{
		#region common or UI Properties
		public Atom atom;
		public EBATCore core;
		public bool disableLog;
		#endregion
		
		public Guess(EBATCore core){
			atom = core.atom;
			this.core = core;
		}
		
		public void loadSettings(){
			try{
				setControllersAndRigidbodies();
				core.customUnityAssetItem.summonAtomOnly("Trunk", "chestControl", false);
				core.customUnityAssetItem.summonAtomOnly("TrunkLow", "chestControl", false);
				core.customUnityAssetItem.summonAtomOnly("Floor", "control", false);
				/*
				core.customUnityAssetItem.summonCUA("Trunk", "Custom/Assets/vam_gizmo.assetbundle", "assets/prefabs/vam_gizmo.prefab", "chestControl", 0f, 0f, 0f, true, true, false);
				core.customUnityAssetItem.summonCUA("TrunkLow", "Custom/Assets/vam_gizmo.assetbundle", "assets/prefabs/vam_gizmo.prefab", "pelvisControl", 0f, 0f, 0f, true, true, false);
				core.customUnityAssetItem.summonCUA("Floor", "Custom/Assets/vam_gizmo.assetbundle", "assets/prefabs/vam_gizmo.prefab", "control", 0f, 0f, 0f, true, true, false);
				*/
				
			}catch (Exception e){
				core.loggerE(e);
			}

		}
		
		public void CheckFloor(){
			if(core.floor == null){
				SetFloor();
			}
		}
		
		public void CheckTrunk(){
			if(core.trunk == null){
				SetTrunk();
			}
			if(core.trunk == null){
				return;
			}
			core.trunk.transform.position = chestR.position;
			core.trunk.transform.rotation = GetTrunkRotation("chestControl", "neckControl", "lShoulderControl", "rShoulderControl");
		}
		
		public void CheckTrunkLow(){
			if(core.trunkLow == null){
				SetTrunkLow();
			}
			core.trunkLow.transform.rotation = GetTrunkRotation("pelvisControl", "neckControl", "lThighControl", "rThighControl");
		}
		
		public void SetFloor(){
			if(core.floor == null){
				core.floor = core.customUnityAssetItem.GetCUAAtom("Floor");
			}
			
			if(core.floor == null){
				core.customUnityAssetItem.PrintAllAtomId();
			}
		}
		
		public void SetTrunk(){
			if(core.trunk == null){
				core.trunk = core.customUnityAssetItem.GetCUAAtom("Trunk");
			}
			
			if(core.trunk == null){
				core.customUnityAssetItem.PrintAllAtomId();
			}else{
				core.trunk.transform.SetPositionAndRotation(Utility.getPosition(chestC), Utility.getRotation(chestC));
				core.trunk.transform.rotation = GetTrunkRotation("chestControl", "neckControl", "lShoulderControl", "rShoulderControl");
			}
		}
		
		public void SetTrunkLow(){
			if(core.trunkLow == null){
				core.trunkLow = core.customUnityAssetItem.GetCUAAtom("TrunkLow");
			}
			if(core.trunkLow == null){
				core.customUnityAssetItem.PrintAllAtomId();
			}else{
				core.trunkLow.transform.SetPositionAndRotation(Utility.getPosition(pelvisC), Utility.getRotation(pelvisC));
				core.trunkLow.transform.rotation = GetTrunkRotation("pelvisControl", "neckControl", "lThighControl", "rThighControl");
			}
		}
		
		public Quaternion GetTrunkRotation(string down, string up, string left, string right){
			Vector3 directionUp = Utility.GetDirection(core.atom, down, up);
			Vector3 directionRight = Utility.GetDirection(core.atom, left, right);
			Quaternion rot = Quaternion.LookRotation(directionRight, directionUp);
			return Quaternion.AngleAxis(-90f, directionUp) * rot;
		}
		
		public void testAll(){
			try{
				CheckFloor();
				CheckTrunk();
				LegsState();
				BodyState();
//				FaceAngle();
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		public void BodyState(){
			CheckFloor();
			CheckTrunk();
			float faceUp;
			float stand;
			float faceDown;
			Transform target;
			target = core.trunk.transform;
			
			faceUp   = _angle(core.floor.transform, target.transform, target.transform.right, -90);
			stand    = _angle(core.floor.transform, target.transform, target.transform.right, 0);
			faceDown = _angle(core.floor.transform, target.transform, target.transform.right, 90);
			core.debug("BodyState:faceUp/stand/faceDown:" + faceUp + "/" + stand + "/" + faceDown);			
		}
		
		
		public void LegsState(){
			getThighsBendedStage();
			getThighsOpenedStage();
		}
		
		public int getThighsOpenedStage(){
			int result = 400;
			Vector3 pos1 = lShinR.position - pelvisR.position;
			Vector3 pos2 = rShinR.position - pelvisR.position;
			float angleV3 = Vector3.Angle(pos1, pos2);

			if(angleV3 > 60){
				result = 100;
			}else if(angleV3 > 45){
				result = 200;
			}else if(angleV3 > 0){
				result = 300;
			}
			
			return result;
		}
		
		public int getResult(int var1, int var2){
			int result = 0;
			
			if(var1 == 1 && var2 == 1){
				result = 100;
			}else if(var1 <= 2 && var2 <= 2){
				if(var1 == var2){
					result = 200;
				}else{
					result = 200 + var1 * 10 + var2;
				}
			}else if(var1 <= 3 && var2 <= 3){
				if(var1 == var2){
					result = 300;
				}else{
					result = 300 + var1 * 10 + var2;
				}
			}else{
				result = 400;
				result += var1 * 10 + var2;
			}
			core.debug("getResult: result: " + result);
			return result;
		}
		
		public int getThighsBendedStage(){
			int result = 0;
			int leftBended = getThighBendedStage("left");
			int rightBended = getThighBendedStage("right");
			return getResult(leftBended, rightBended);
		}
		
		public int getThighBendedStage(string leg = "left"){
			CheckTrunkLow();
			float full;
			float half;
			float zero;
			float back;
			Transform target;

			target = (leg == "left") ? lThighC.transform : rThighC.transform;
			full = _angle(core.trunkLow.transform, target.transform, target.transform.right, -90);
			half = _angle(core.trunkLow.transform, target.transform, target.transform.right, -45);
			zero = _angle(core.trunkLow.transform, target.transform, target.transform.right, 0);
			back = _angle(core.trunkLow.transform, target.transform, target.transform.right, 15);
				
			int num = getMinIndex(full, half, zero, back);
//			core.debug("getThighBendedStage(" + leg + "):full/half/zero/back:" + full + "/" + half + "/" + zero + "/" + back);
//			core.debug("getThighBendedStage(" + leg + "):" + num);
			return num;
		}
		
		public int getMinIndex(float num1 = 999f, float num2 = 999f, float num3 = 999f, float num4 = 999f, float num5 = 999f){
			float[] numbers = {num1, num2, num3, num4, num5};
			float min = num1;
			int index = 1;
			for(int i = 1; i < 5; i++){
				if(min > numbers[i]){
					min = num1;
					index = i + 1;
				}
			}
			return index;
		}
						
		public int GetFaceAngleHorizontal(){
			CheckTrunk();
			return AngleStage(false, core.trunk.transform, head, 20f, 50f, 0f);
		}
		
		public int GetFaceAngleVertical(){
			CheckTrunk();
			return AngleStage(true, core.trunk.transform, head, 20f, 40f,-10f);
		}
		
		public void FaceAngle(){
			try{
				int ud = AngleStage(true,  core.trunk.transform, head, 20f, 40f, -10f);
				int lr = AngleStage(false, core.trunk.transform, head, 20f, 50f, 0f);
				string direction = "V" + ud  + "H" + lr;
				
				Vector3 axis = neck.position - chest.position;
				Quaternion testRot = Quaternion.AngleAxis(90, axis) * chest.rotation;
				eyes.rotation = testRot;
				
				core.debug("direction:" + direction);
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void test1(){
		}

		public void test2(){
		}

		public void test3(){
		}

		public int AngleStage(bool isVertical, Transform baseT, Transform target, float threashold1 = 20f, float threshold2 = 50f, float correction = 0f){
			CheckTrunk();
			Vector3 axis = new Vector3(0f, 0f, 0f);
			int degree = 90;
			axis = (isVertical) ? baseT.transform.right : baseT.transform.up;
			float angleStart   = _angle(baseT, target, axis,  degree);
			float angleEnd     = _angle(baseT, target, axis, -1 * degree);
			float diff = angleStart - angleEnd;			
			float measure = diff + correction;
			
			if(Mathf.Abs(measure) < threashold1){
				return 3;
			}else{
				if(Mathf.Abs(measure) < threshold2){
					return (measure > 0) ? 2 : 4;
				}else{
					return (measure > 0) ? 1 : 5;
				}
			}
			return 1;
		}	

		public float _angle(Transform baseT, Transform target, Vector3 axis, int degree){
			Quaternion standard = Quaternion.AngleAxis(degree, axis) * baseT.rotation;
			float angle = Quaternion.Angle(standard, target.rotation);
			return angle;
		}
		
		public void PrintRotation(){
			CheckTrunk();
			if(core.trunk == null){
				core.debug("trunk is null!!");
				return;
			}
			
			FreeControllerV3 fcHead = atom.GetStorableByID("headControl") as FreeControllerV3;
			Quaternion rotHead = fcHead.transform.rotation;
			Quaternion rotTrunk = core.trunk.transform.rotation;
			Quaternion rotDiff = Quaternion.RotateTowards(rotTrunk, rotHead, 1.0f);

			Vector3 angle = rotDiff.eulerAngles;
			core.debug("rotDiff: " + rotDiff.ToString("F3"));
			
			float fAngleUp = Vector3.Angle(core.trunk.transform.up, fcHead.transform.up);
			core.debug("angleHead: " + fcHead.transform.rotation.eulerAngles.ToString("F0"));
			core.debug("angleTrunk: " + core.trunk.transform.rotation.eulerAngles.ToString("F0"));
			core.debug("angleDiff: " + angle.ToString("F0"));
			core.debug("fAngleUp: " + fAngleUp);
		}
		
		public void SyncHeadNeckTrunkRotation(){
			CheckTrunk();
			if(core.trunk == null){
				return;
			}
			
			FreeControllerV3 fcHead = atom.GetStorableByID("headControl") as FreeControllerV3;
			FreeControllerV3 fcNeck = atom.GetStorableByID("neckControl") as FreeControllerV3;
			
			fcHead.transform.rotation = core.trunk.transform.rotation;
			fcNeck.transform.rotation = core.trunk.transform.rotation;			
		}
		
		public void PrintPositionsAndDistance(string type = "all", string pivot = "", float fluctuation = 0.01f){
			string[] namesAll = {
				"headControl",
				"neckControl",
				"chestControl",
				"hipControl",
				"pelvisControl",
				"rElbowControl",
				"lElbowControl",
				"rHandControl",
				"lHandControl",
				"rKneeControl",
				"lKneeControl",
				"rFootControl",
				"lFootControl",
				"rThighControl",
				"lThighControl",
				"rArmControl",
				"lArmControl",				
			};
			
			string[] namesArms = {
//				"chestControl",
				"lHandControl",
				"rHandControl",
				"lElbowControl",
				"rElbowControl",
//				"rArmControl",
//				"lArmControl",				
			};
			
			string[] namesLegs = {
//				"chestControl",
//				"hipControl",
//				"pelvisControl",
				"rKneeControl",
				"lKneeControl",
				"rFootControl",
				"lFootControl",
//				"rThighControl",
//				"lThighControl",
			};
			
			string[] names;
			if(type == "all"){
				names = namesAll;
			}else{
				names = (type == "arms") ? namesArms : namesLegs;
			}
			Vector3 pivotPos = new Vector3(0f, 0f, 0f);
			if(pivot != ""){
				FreeControllerV3 fc = atom.GetStorableByID(pivot) as FreeControllerV3;
				pivotPos = fc.transform.position;
			}
			core.debug("REPORT");
			core.debug("-----");
			string command = "";
			foreach(var name in names){
				FreeControllerV3 fc = atom.GetStorableByID(name) as FreeControllerV3;
				Vector3 pos = fc.transform.position;
				Quaternion rot = fc.transform.rotation;
				if(pivot != ""){
					Vector3 diff = pos - pivotPos;
					core.debug("    " + name + ": position: " + pos + ": pivotPos("  + pivot + "): " + round(diff.y, 2) + ", " + round(diff.z, 2) + ", " + round(diff.x, 2));
					string sname1 = name.Replace("Control", "");
					string sname2 = pivot.Replace("Control", "");
					command += "TR:" + sname1 + ":" + sname2 + ":" + round(diff.y, 2) + ":" + round(diff.z, 2) + ":" + round(diff.x, 2) + ":" + fluctuation + ",";
				}else{
					core.debug("    " + name + ": position: " + pos);
				}
			}
			command = command.Substring(0, command.Length - 1);
			core.debug("command: " + command);
			core.debug("-----");
		}
		
		public float round(float value, int keta){
			return (float)Math.Round(value, keta,  MidpointRounding.AwayFromZero);
		}
		
		public void setControllersAndRigidbodies(){
			try{
				
			controlC = atom.GetStorableByID("control"   ) as FreeControllerV3;
			hipC   = atom.GetStorableByID("hipControl"   ) as FreeControllerV3;
			eyesC  = atom.GetStorableByID("eyeTargetControl"   ) as FreeControllerV3;
			headC  = atom.GetStorableByID("headControl"  ) as FreeControllerV3;
			neckC  = atom.GetStorableByID("neckControl"  ) as FreeControllerV3;
			rArmC  = atom.GetStorableByID("rArmControl"  ) as FreeControllerV3;
			lArmC  = atom.GetStorableByID("lArmControl"  ) as FreeControllerV3;
			chestC = atom.GetStorableByID("chestControl" ) as FreeControllerV3;
			rKneeC = atom.GetStorableByID("rKneeControl" ) as FreeControllerV3;
			lKneeC = atom.GetStorableByID("lKneeControl" ) as FreeControllerV3;
			rHandC = atom.GetStorableByID("rHandControl" ) as FreeControllerV3;
			lHandC = atom.GetStorableByID("lHandControl" ) as FreeControllerV3;
			rFootC = atom.GetStorableByID("rFootControl" ) as FreeControllerV3;
			lFootC = atom.GetStorableByID("lFootControl" ) as FreeControllerV3;
			rThighC= atom.GetStorableByID("rThighControl") as FreeControllerV3;
			lThighC= atom.GetStorableByID("lThighControl") as FreeControllerV3;
			pelvisC= atom.GetStorableByID("pelvisControl") as FreeControllerV3;
			rElbowC= atom.GetStorableByID("rElbowControl") as FreeControllerV3;
			lElbowC= atom.GetStorableByID("lElbowControl") as FreeControllerV3;
			rShoulderC= atom.GetStorableByID("rShoulderControl") as FreeControllerV3;
			lShoulderC= atom.GetStorableByID("lShoulderControl") as FreeControllerV3;
			abdomen2C= atom.GetStorableByID("abdomen2Control") as FreeControllerV3;
			abdomenC= atom.GetStorableByID("abdomenControl") as FreeControllerV3;
			
			control= controlC.transform   ;
			hip    = hipC.transform   ;
			eyes   = eyesC.transform   ;
			head   = headC.transform  ;
			neck   = neckC.transform  ;
			rArm   = rArmC.transform  ;
			lArm   = lArmC.transform  ;
			chest  = chestC.transform ;
			rKnee  = rKneeC.transform ;
			lKnee  = lKneeC.transform ;
			rHand  = rHandC.transform ;
			lHand  = lHandC.transform ;
			rFoot  = rFootC.transform ;
			lFoot  = lFootC.transform ;
			rThigh = rThighC.transform;
			lThigh = lThighC.transform;
			pelvis = pelvisC.transform;
			rElbow = rElbowC.transform;
			lElbow = lElbowC.transform;
			rShoulder = rShoulderC.transform;
			lShoulder = lShoulderC.transform;
			abdomen2  = abdomen2C.transform;
			abdomen   = abdomenC.transform;
			
			controlR = Utility.findRigidBody(core.atom, "control");
			hipR = Utility.findRigidBody(core.atom, "hip");
			pelvisR = Utility.findRigidBody(core.atom, "pelvis");
			abdomenR = Utility.findRigidBody(core.atom, "abdomen");
			abdomen2R = Utility.findRigidBody(core.atom, "abdomen2");
			chestR = Utility.findRigidBody(core.atom, "chest");
			neckR = Utility.findRigidBody(core.atom, "neck");
			headR = Utility.findRigidBody(core.atom, "head");
			rThighR = Utility.findRigidBody(core.atom, "rThigh");
			rShinR = Utility.findRigidBody(core.atom, "rShin");
			rFootR = Utility.findRigidBody(core.atom, "rFoot");
			lThighR = Utility.findRigidBody(core.atom, "lThigh");
			lShinR = Utility.findRigidBody(core.atom, "lShin");
			lFootR = Utility.findRigidBody(core.atom, "lFoot");
			lPectoralR = Utility.findRigidBody(core.atom, "lPectoral");
			lNippleR = Utility.findRigidBody(core.atom, "lNipple");
			rPectoralR = Utility.findRigidBody(core.atom, "rPectoral");
			rNippleR = Utility.findRigidBody(core.atom, "rNipple");
			rCollarR = Utility.findRigidBody(core.atom, "rCollar");
			rShldrR = Utility.findRigidBody(core.atom, "rShldr");
			rForeArmR = Utility.findRigidBody(core.atom, "rForeArm");
			rHandR = Utility.findRigidBody(core.atom, "rHand");
			lCollarR = Utility.findRigidBody(core.atom, "lCollar");
			lShldrR = Utility.findRigidBody(core.atom, "lShldr");
			lForeArmR = Utility.findRigidBody(core.atom, "lForeArm");
			lHandR = Utility.findRigidBody(core.atom, "lHand");
			}catch(Exception e){
				core.loggerE(e);
			}
		}
				
		#region fc
		protected FreeControllerV3 controlC;
		protected FreeControllerV3 eyesC;
		protected FreeControllerV3 headC;
		protected FreeControllerV3 neckC;
		protected FreeControllerV3 chestC;
		protected FreeControllerV3 hipC;
		protected FreeControllerV3 rThighC;
		protected FreeControllerV3 lThighC;
		protected FreeControllerV3 rKneeC;
		protected FreeControllerV3 lKneeC;
		protected FreeControllerV3 pelvisC;
		protected FreeControllerV3 rArmC;
		protected FreeControllerV3 lArmC;
		protected FreeControllerV3 rElbowC;
		protected FreeControllerV3 lElbowC;
		protected FreeControllerV3 rHandC;
		protected FreeControllerV3 lHandC;
		protected FreeControllerV3 rFootC;
		protected FreeControllerV3 lFootC;
		protected FreeControllerV3 rShoulderC;
		protected FreeControllerV3 lShoulderC;
		protected FreeControllerV3 abdomen2C;
		protected FreeControllerV3 abdomenC;
		
		protected Transform control;
		protected Transform eyes;
		protected Transform head;
		protected Transform neck;
		protected Transform chest;
		protected Transform hip;
		protected Transform rThigh;
		protected Transform lThigh;
		protected Transform rKnee;
		protected Transform lKnee;
		protected Transform pelvis;
		protected Transform rArm;
		protected Transform lArm;
		protected Transform rElbow;
		protected Transform lElbow;
		protected Transform rHand;
		protected Transform lHand;
		protected Transform rFoot;
		protected Transform lFoot;
		protected Transform rShoulder;
		protected Transform lShoulder;
		protected Transform abdomen2;
		protected Transform abdomen;
		
		protected Rigidbody controlR;
		protected Rigidbody hipR;
		protected Rigidbody pelvisR;
		protected Rigidbody abdomenR;
		protected Rigidbody abdomen2R;
		protected Rigidbody chestR;
		protected Rigidbody neckR;
		protected Rigidbody headR;
		protected Rigidbody rThighR;
		protected Rigidbody rShinR;
		protected Rigidbody rFootR;
		protected Rigidbody lThighR;
		protected Rigidbody lShinR;
		protected Rigidbody lFootR;
		protected Rigidbody lPectoralR;
		protected Rigidbody lNippleR;
		protected Rigidbody rPectoralR;
		protected Rigidbody rNippleR;
		protected Rigidbody rCollarR;
		protected Rigidbody rShldrR;
		protected Rigidbody rForeArmR;
		protected Rigidbody rHandR;
		protected Rigidbody lCollarR;
		protected Rigidbody lShldrR;
		protected Rigidbody lForeArmR;
		protected Rigidbody lHandR;
		#endregion
	}
}