#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class ActionAndRatio{
		public Dictionary<string, float> actionAndRatioMap;
		public List<string> actionList;
		public List<float> frequencyBin;
		
		public ActionAndRatio(){
			actionAndRatioMap = new Dictionary<string, float>();
			actionList = new List<string>();
			frequencyBin = new List<float>();
		}

		public void InitActionList(List<string> actionList){
			foreach(var name in actionList){
				this.actionList.Add(name);
			}
		}
		
		public void SetActionAndRatio(string actionName, float ratio){
			if(actionAndRatioMap.ContainsKey(actionName)){
				actionAndRatioMap[actionName] = ratio;
			}else{
				actionAndRatioMap.Add(actionName, ratio);
			}
		}
		
		public float GetActionAndRatio(string actionName){
			if(actionAndRatioMap.ContainsKey(actionName)){
				return actionAndRatioMap[actionName];
			}else{
				return 0f;
			}
		}
		
		public void MakeRatioTable(){
			float current = 0;
			frequencyBin.Clear();
			frequencyBin.Add(0f);
			foreach(var name in actionList){
				current += actionAndRatioMap[name];
				frequencyBin.Add(current);
			}
			for(int i = 0; i < frequencyBin.Count; i++){
				frequencyBin[i] = frequencyBin[i] / current;
			}
		}
		
		public string PickAction(){
			float rnd = UnityEngine.Random.Range(0.0f,1.0f);
			int num = 0;
			for(int i = 0; i < frequencyBin.Count; i++){
				if(rnd < frequencyBin[i]){
					num = i - 1;
					break;
				}
			}
			return actionList[num];
		}
	}
	
	public class StageAndEvent{
		public List<string> actionList;
		public Dictionary<string, float> fireRatioMap;
		public Dictionary<string, ActionAndRatio> actionAndRatioMap;
		
		public StageAndEvent(){
			actionList = new List<string>();
			fireRatioMap = new Dictionary<string, float>();
			actionAndRatioMap = new Dictionary<string, ActionAndRatio>();
		}
		
		public void InitActionList(string[] actionList){
			foreach(var name in actionList){
				this.actionList.Add(name);
			}
		}
		
		public List<string> ListStages(){
			List<string> list = new List<string>();
			foreach(KeyValuePair<string, ActionAndRatio> kvp in actionAndRatioMap){
				list.Add(kvp.Key);
			}
			return list;
		}
		
		public void InitFireRatio(Dictionary<string, float> fireRatio){
			foreach(KeyValuePair<string, float> kvp in fireRatio){
				this.fireRatioMap.Add(kvp.Key, kvp.Value);
			}
		}
		
		public void SetFireRatio(string stage, float ratio){
			if(fireRatioMap.ContainsKey(stage)){
				fireRatioMap[stage] = ratio;
			}else{
				fireRatioMap.Add(stage, ratio);
			}
		}
		
		public float GetFireRatio(int stage, int count){
			string name = ResolveStageName(stage, count);
			return fireRatioMap[name];
		}

		public float GetFireRatio(string stage){
			return fireRatioMap[stage];
		}
		
		public string PickAction(string stage, bool bForce = false){
			float rnd = UnityEngine.Random.Range(0.0f,1.0f);
			if(rnd < fireRatioMap[stage] || bForce){
				return actionAndRatioMap[stage].PickAction();
			}else{
				return "";
			}
		}
		
		public string PickAction(int stage, int count, bool bForce = false){
			string name = ResolveStageName(stage, count);
			return PickAction(name, bForce);
		}
		
		public void InitActionAndRatioByStage(string stage, Dictionary<string, float> actionAndRatio){
			if(!this.actionAndRatioMap.ContainsKey(stage)){
				this.actionAndRatioMap.Add(stage, new ActionAndRatio());
				this.actionAndRatioMap[stage].InitActionList(this.actionList);
			}
			foreach(KeyValuePair<string, float> kvp in actionAndRatio){
				this.actionAndRatioMap[stage].SetActionAndRatio(kvp.Key, kvp.Value);
			}
			this.actionAndRatioMap[stage].MakeRatioTable();
		}
		
		public void SetActionAndRatio(string stage, string actionName, float ratio){
			this.actionAndRatioMap[stage].SetActionAndRatio(actionName, ratio);
			this.actionAndRatioMap[stage].MakeRatioTable();
		}
		
		public float GetRatioByStage(string stage, string actionName){
			return actionAndRatioMap[stage].GetActionAndRatio(actionName);
		}
		
		public string ResolveStageName(int stage, int count){
			string name = (stage == 5) ? "Stage=5" : "Stage<" + count;
			if(stage == 5){
				int threshold = 99;
				if(count < 1){ threshold = 1; }
				else if(count < 3){ threshold = 3; }
				else if(count < 6){ threshold = 6; }
				else if(count < 9){ threshold = 9; }
				name += ":OrgasmCount<" + ((count == 99) ? threshold.ToString() : ("0" + threshold.ToString()));
			}else{
				name += ":OrgasmCount<99";
			}
			return name;
		}
	}
	
	public class AxisAndAngles{
		public Dictionary<string, float> angleMap;
		
		public AxisAndAngles(){
			angleMap = new Dictionary<string, float>(){
				{"up", 0f},
				{"forward", 0f},
				{"right", 0f},
				{"forward2", 0f},
				{"right2", 0f},
			};
		}
		
		public void SetAngle(string axis, float degree){
			if(angleMap.ContainsKey(axis)){
				angleMap[axis] = degree;
			}else{
				angleMap.Add(axis, degree);
			}
		}
		
		public float GetAngle(string axis){
			if(angleMap.ContainsKey(axis)){
				return angleMap[axis];
			}else{
				return 0f;
			}
		}		
	}
	
	public class HeadController
	{
		#region common or UI Properties
		public Atom atom;
		public EBATCore core;
		public bool disableLog;
		public bool headMotionOnGlobal = true;
		#endregion
		
		#region bellyControl
		protected Dictionary<string, Quaternion> initialQuat;
		protected Dictionary<string, Quaternion> recoveryQuat;
		protected Dictionary<string, Quaternion> targetQuat;
		protected Dictionary<string, float> progress;
		protected Dictionary<string, CycleManager> timing;
		protected Dictionary<string, FreeControllerV3> ctrl;
		protected Dictionary<string, ControllerBackup> backup;
		protected Dictionary<string, bool> isActive;
		protected Dictionary<string, bool> isForce;
		
		protected Dictionary<string, float> lastActionTime;
		protected float actionIntervalRatioMin = 0.5f;
		protected float actionIntervalRatioMax = 2.0f;
		
		protected string[] keyHead = {
			"headControl"
		};
		
		protected string[] actionList = {
			"Turn Head Back 1",
			"Turn Head Back 2",
			"Glance",
			"Face Forward",
			"Face Sideways 1",
			"Face Sideways 2",
			"Face Sideways 3",
		};
		
		protected string[] eventList = {
			"Stage<3:OrgasmCount<99",
			"Stage<5:OrgasmCount<99",
			"Stage=5:OrgasmCount<01",
			"Stage=5:OrgasmCount<03",
			"Stage=5:OrgasmCount<06",
			"Stage=5:OrgasmCount<09",
			"Stage=5:OrgasmCount<99",
		};
		
		protected string[] axisList = {
			"up", "forward", "right", "forward2", "right2",
		};
		
		protected string[] keyCommand = {
			"Turn Head Back 1"
			, "Turn Head Back 2"
			, "Face Sideways 1"
			, "Face Sideways 2"
			, "Face Sideways 3"
			, "Glance"
			, "Face Forward", "faceForwardSlowly",
		};
		
		protected FreeControllerV3 headControl;
		protected FreeControllerV3 chestControl;
		protected FreeControllerV3 neckControl;
		
		protected Dictionary<string, float> actionInterval = new Dictionary<string, float>
		{
			["Turn Head Back 1"] = 0.5f
			, ["Turn Head Back 2"] = 6.0f
			, ["Face Sideways 1"] = 4.0f
			, ["Face Sideways 2"] = 6.0f
			, ["Face Sideways 3"] = 8.0f
			, ["Glance"] = 4.0f
			, ["Face Forward"] = 3.0f
			, ["faceForwardSlowly"] = 3.0f
		};
		
		public Dictionary<string, float> targetStrength = new Dictionary<string, float>
		{
			["headControl"]       = -10
		};

		public Dictionary<string, float> closingTimeOffset1 = new Dictionary<string, float>
		{
			["Turn Head Back 1"] = 0.1f
			, ["Turn Head Back 2"] = 0.2f
			, ["Face Sideways 1"] = 0.3f
			, ["Face Sideways 2"] = 0.5f
			, ["Face Sideways 3"] = 0.5f
			, ["Glance"] = 0.4f
			, ["Face Forward"] = 0.2f
			, ["faceForwardSlowly"] = 5.0f
		};
		public Dictionary<string, float> closingTimeOffset2 = new Dictionary<string, float>
		{
			["Turn Head Back 1"] = 0.5f
			, ["Turn Head Back 2"] = 1.5f
			, ["Face Sideways 1"] = 0.3f
			, ["Face Sideways 2"] = 0.5f
			, ["Face Sideways 3"] = 1.0f
			, ["Glance"] = 0.4f
			, ["Face Forward"] = 0.2f
			, ["faceForwardSlowly"] = 5.0f
		};		
		
		protected Dictionary<string, AxisAndAngles> angleMap;
//		protected Dictionary<string, RatioAndStage> ratioMap;
		protected StageAndEvent stageAndEventMap;
		#endregion
		
		#region breath
		#endregion
		
		#region head
		public float bendBackwardThreshold = 0.2f;
		public float bendBackwardStrongThreshold = 0.2f;
		public float rbendBackward = 0.3f;
		public float rBigMotion = 0.03f;
		public float rFaceSidewaysWeak = 0.2f;
		public float rFaceSideways     = 0.2f;
		public float rFaceSidewaysPeak = 0.2f;
		public float glanceThreshold = 0.2f;
		
		public float timeRandomMinRatio = 0.7f;
		public float timeRandomMaxRatio = 2.0f;
		#endregion
		
		private float lastTime;
		
		public HeadController(EBATCore core){
			atom = core.atom;
			this.core = core;
			headControl  = Utility.getFreeControllerV3(atom, "headControl");
			neckControl  = Utility.getFreeControllerV3(atom, "neckControl");
			chestControl = Utility.getFreeControllerV3(atom, "chestControl");
			
		}
		
		public void loadSettings(){
//			try{				
				headMotionOnGlobal = core.settings["General"]["Head Motion On"].AsBool;

				angleMap = new Dictionary<string, AxisAndAngles>();
				foreach(var actionName in actionList){
					angleMap.Add(actionName, new AxisAndAngles());
				}
				updateAngle();
				
				stageAndEventMap = new StageAndEvent();
				stageAndEventMap.InitActionList(actionList);
				
				stageAndEventMap.InitFireRatio(MakeDictionary("fireRatio"));
				foreach(var stageName in eventList){
					stageAndEventMap.InitActionAndRatioByStage(stageName, MakeDictionary("ratioByStage", stageStr:stageName));
				}
				/*
			}catch(Exception e){
				core.loggerE(e);
			}
			*/
		}
				
		public Dictionary<string, float> MakeDictionary(string type = "fireRatio", string stageStr = ""){
			Dictionary<string, float> tmpDic = new Dictionary<string, float>();
			if(type == "fireRatio"){
				foreach(var stageName in eventList){
					tmpDic.Add(stageName, core.settings["HeadControlRatio"]["EventFireRatio"][stageName].AsFloat);
				}
			}else if(type == "ratioByStage"){
				foreach(var actionName in actionList){
					tmpDic.Add(actionName, core.settings["HeadControlRatio"][stageStr][actionName].AsFloat);
				}				
			}
			return tmpDic;
		}
		
		public List<string> ListStages(){
			return stageAndEventMap.ListStages();
		}
		
		public string[] ListActions(){
			return actionList;
		}
		
		public float GetAngle(string actionName, string axis){
			return angleMap[actionName].GetAngle(axis);
		}
		
		public float GetFireRatio(int stage, int count){
			return stageAndEventMap.GetFireRatio(stage, count);
		}
		
		public float GetFireRatio(string stage){
			return stageAndEventMap.GetFireRatio(stage);
		}
		
		public float GetRatioByStage(string stage, string actionName){
			return stageAndEventMap.GetRatioByStage(stage, actionName);
		}
		
		public void updateAngle(){
			foreach(var actionName in actionList){
				foreach(var axis in axisList){
					angleMap[actionName].SetAngle(axis, core.settings["HeadControlAngle"][actionName][axis].AsFloat);
				}
			}
		}
		
		public void updateRatio(){
			/*
			foreach(var actionName in actionList){
				foreach(var stage in ratioList){
					ratioMap[actionName].SetRatio(stage, core.settings["HeadControlRatio"][stage][actionName].AsFloat);
				}
			}			
			*/
		}
		
		public void applySettings(){
			setControllerAndProgress();
		}
		
		public void maintenance(){

		}

		#region orgasm functions
		public void setControllerAndProgress(){
			try{
				ctrl   = new Dictionary<string, FreeControllerV3>();
				backup = new Dictionary<string, ControllerBackup>();
				progress = new Dictionary<string, float>();
				lastActionTime = new Dictionary<string, float>();
				timing   = new Dictionary<string, CycleManager>();
				isActive = new Dictionary<string, bool>();
				isForce = new Dictionary<string, bool>();

				initialQuat = new Dictionary<string, Quaternion>();
				recoveryQuat = new Dictionary<string, Quaternion>();
				targetQuat = new Dictionary<string, Quaternion>();
				
				for(int i = 0; i < keyHead.Length; i++){
					ctrl[keyHead[i]] = atom.GetStorableByID(keyHead[i]) as FreeControllerV3;
					backup.Add(keyHead[i], new ControllerBackup());
					backup[keyHead[i]].backUp(ctrl[keyHead[i]]);
					progress[keyHead[i]] = 0;
				}
				
				foreach(var key in keyCommand){
					isActive[key] = false;
					isForce[key] = false;
					timing.Add(key, new CycleManager());
					lastActionTime.Add(key, 0.0f);
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		public void storePositionAndRotation(){
			try{
				initialQuat = new Dictionary<string, Quaternion>();
				
				for(int i = 0; i < keyHead.Length; i++){
					backup[keyHead[i]].backUp(ctrl[keyHead[i]]);
					recoveryQuat[keyHead[i]] = ctrl[keyHead[i]].transform.rotation;
					initialQuat[keyHead[i]] = ctrl[keyHead[i]].transform.rotation;
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void changePositionAndRotationState(){
			try{
				for(int i = 0; i < keyHead.Length; i++){
					ctrl[keyHead[i]].SetRotationStateFromString("On");
					ctrl[keyHead[i]].SetPositionStateFromString("On");
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}		
		
		public void restoreStatus(){
			try{
				for(int i = 0; i < keyHead.Length; i++){
					ctrl[keyHead[i]] = backup[keyHead[i]].recoveryStatus(ctrl[keyHead[i]]);
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void restorePosition(){
			try{
				for(int i = 0; i < keyHead.Length; i++){
					ctrl[keyHead[i]] = backup[keyHead[i]].recoveryPosition(ctrl[keyHead[i]]);
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void restoreRotation(){
			try{
				for(int i = 0; i < keyHead.Length; i++){
					ctrl[keyHead[i]] = backup[keyHead[i]].recoveryRotation(ctrl[keyHead[i]]);
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void restorePositionAndRotation(){
			try{
				for(int i = 0; i < keyHead.Length; i++){
					ctrl[keyHead[i]] = backup[keyHead[i]].recovery(ctrl[keyHead[i]]);
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public bool isFacingForward(){
			try{
				return (core.guess.GetFaceAngleHorizontal() == 3);
			}catch(Exception e){
				core.loggerE(e);
				return false;
			}
		}
		
		public float rand(string type = "small"){
			if(type == "big"){
				return UnityEngine.Random.Range(0.5f,3.0f);				
			}else{
				return UnityEngine.Random.Range(0.8f,1.2f);				
			}
		}

		public bool readyNextAction(string type){
			bool result = true;
			if(isActive["Face Sideways 1"] 
				|| isActive["Face Sideways 2"] 
				|| isActive["Face Sideways 3"] 
				|| isActive["Turn Head Back 2"] 
				|| isActive["Glance"]){
					result = false;
			}
			float offsetRatio = UnityEngine.Random.Range(actionIntervalRatioMin, actionIntervalRatioMax);
			if(core.currentTime < lastActionTime[type] + actionInterval[type] * offsetRatio){
				result = false;
			}
			
			return result;
		}
		
		public void updateMotion(string type, bool bForce = false, float strength = 0f){
			try{
				if(false){
//				if(isActive[type]){
					return;
				}else{
					if(!bForce && readyNextAction(type) == false){
						return;
					}
					float time = core.currentTime;
					core.guess.CheckTrunk();
					float cFacingForward = isFacingForward() ? 0.5f : 1.0f;
					float time1 = time + closingTimeOffset1[type] * UnityEngine.Random.Range(timeRandomMinRatio, timeRandomMaxRatio) * AdjustInterval(bForce) * cFacingForward;
					float time2 = time1 + closingTimeOffset2[type] * UnityEngine.Random.Range(timeRandomMinRatio, timeRandomMaxRatio) * AdjustInterval(bForce) * cFacingForward;
					
					lastActionTime[type] = time;
					timing[type].setTimer(time, time1, time2);

					isForce[type] = bForce;
					initialQuat["headControl"] = headControl.transform.rotation;							
					if(isActive[type] == false){
						isActive[type] = true;
						recoveryQuat["headControl"] = headControl.transform.rotation;
					}
					if(type == "Turn Head Back 1"){
						targetQuat["headControl"] = GetBendingBackRotation(bForce);
					}else if(type == "Turn Head Back 2"){
						recoveryQuat["headControl"] = GetBendingBackStrongRotationInit(bForce);
//						initialQuat["headControl"] = GetBendingBackStrongRotationInit();
						targetQuat["headControl"] = GetBendingBackStrongRotationTarget(bForce);
					}else if(type == "Face Forward" || type == "faceForwardSlowly"){
						recoveryQuat["headControl"] = GetBendingBackStrongRotationInit(bForce, type:"Face Forward");
//						initialQuat["headControl"] = GetBendingBackStrongRotationInit();
						targetQuat["headControl"] = GetBendingBackStrongRotationTarget(bForce, type:"Face Forward");
					}else if(type == "Face Sideways 1"){
						recoveryQuat["headControl"] = GetFaceAwayRotation(bForce, fStrength:strength);
						targetQuat["headControl"] = GetFaceAwayRotation(bForce, fStrength:strength);
					}else if(type == "Face Sideways 2"){
						recoveryQuat["headControl"] = GetFaceAwayRotation(bForce, type:"Face Sideways 2");
						targetQuat["headControl"] = GetFaceAwayRotation(bForce, type:"Face Sideways 2");
					}else if(type == "Face Sideways 3"){
						recoveryQuat["headControl"] = GetFaceAwayRotation(bForce, type:"Face Sideways 3");
						targetQuat["headControl"] = GetFaceAwayRotation(bForce, type:"Face Sideways 3");
					}else if(type == "Glance"){
						recoveryQuat["headControl"] = GetFaceAwayRotation(bForce, type:"Glance");
						targetQuat["headControl"] = GetFaceAwayRotation(bForce, type:"Glance");
//						targetQuat["headControl"] = GetGlanceRotation();
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}		
		
		public Quaternion GetBendingBackRotation(bool bForce){
			string name = "Turn Head Back 1";
			float strength  = angleMap[name].angleMap["right"] * rand() * AdjustStrength(bForce);
			return 
				Quaternion.AngleAxis(strength, headControl.transform.right) * 
				headControl.transform.rotation;
		}
		
		public Quaternion GetBendingBackStrongRotationTarget(bool bForce, string type = "strong"){
			string name = "Turn Head Back 2";
			float strength  = angleMap[name].angleMap["right"] * rand() * AdjustStrength(bForce);
			strength = (type == "Face Forward" || type == "faceForwardSlowly") ? 10f : strength;
			return 
				Quaternion.AngleAxis(strength, core.trunk.transform.right) * 
				core.trunk.transform.rotation;
		}
		
		public Quaternion GetBendingBackStrongRotationInit(bool bForce, string type = "strong"){
//			string type = "Turn Head Back 2";
//			float strength  = axisMap[type].angleMap["right"] * rand() * AdjustStrength(bForce);
			float strengthBase = 10f;
			float strength  = 10f * rand() * AdjustStrength(bForce);
			strength = (type == "Face Forward" || type == "faceForwardSlowly") ? 10f : strength;
			return
				Quaternion.AngleAxis(strength, core.trunk.transform.right) * 
				core.trunk.transform.rotation;
		}
		
		public Quaternion GetFaceAwayRotation(bool bForce, string type = "Face Sideways 1", float fStrength = 0f){
			int angleH = core.guess.GetFaceAngleHorizontal();
			int angleV = core.guess.GetFaceAngleVertical();
			
			float sign = 1.0f;
			float degreeUp       =  angleMap[type].angleMap["up"]       * rand();
			float degreeForward  =  angleMap[type].angleMap["forward"]  * rand();
			float degreeForward2 =  angleMap[type].angleMap["forward2"] * rand();
			float degreeRight    =  angleMap[type].angleMap["right"]    * rand();
			float degreeRight2   =  angleMap[type].angleMap["right2"]   * rand();
			
			if(angleH == 3){ //"C"
				sign = (UnityEngine.Random.Range(0.0f,1.0f) > 0.5) ? 1.0f : -1.0f;
			}else{
				sign = (angleH > 3) ? -1.0f : 1.0f; // L:1,2 C:3 R:4,5
			}
			string arrow = (sign == -1.0f) ? "Left" : "Right";
			
			Quaternion rot =
					Quaternion.AngleAxis(degreeRight2, core.trunk.transform.right) *
					Quaternion.AngleAxis(degreeForward2 * sign, core.trunk.transform.forward) *
					Quaternion.AngleAxis(degreeForward * -1.0f * sign, core.trunk.transform.forward) *
					Quaternion.AngleAxis(degreeUp * sign, core.trunk.transform.up) *
					Quaternion.AngleAxis(degreeRight, core.trunk.transform.right) *
					core.trunk.transform.rotation;
			return rot;
		}

		public void calcProgress(float time, string type){
			try{
				progress[type] = timing[type].getProgress(time);
				
				if(timing[type].getState(time) == 0){
					isActive[type] = false;
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public string getMotionType(string type){
			string result = "RoundTrip";
			if(type == "Turn Head Back 1"){
				result = "RoundTrip";
			}else if(type == "Turn Head Back 2"){
				result = "RoundTrip";
			}else if(type == "Face Sideways 1"){
				result = "OneWay";
			}else if(type == "Glance"){
				result = "OneWay";
			}
			return result;
		}
		
		public void changeJoint(float time, string[] keys, string type){
			try{
				if(!timing.ContainsKey(type)){
					core.debug("key:" + type + " not found.");
				}
				int state = timing[type].getState(time);
				for(int i = 0; i < keys.Length; i++){
					if(state == 0 || (state == 2 && getMotionType(type) == "OneWay")){
						
					}else{
						_keyCheck(keys[i], type);
						if(state == 1){
							ctrl[keys[i]].transform.rotation = Quaternion.Slerp(
								initialQuat[keys[i]], targetQuat[keys[i]], progress[type]);						
						}else if(state == 2){
							ctrl[keys[i]].transform.rotation = Quaternion.Slerp(
								recoveryQuat[keys[i]], targetQuat[keys[i]], progress[type]);
						}
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		public void _keyCheck(string key, string type){
			if(!ctrl.ContainsKey(key)){
				core.debug("ctrl key:" + key + " not found.");
			}
			if(!initialQuat.ContainsKey(key)){
				core.debug("initialQuat key:" + key + " not found.");
			}
			if(!targetQuat.ContainsKey(key)){
				core.debug("targetQuat key:" + key + " not found.");
			}
			if(!recoveryQuat.ContainsKey(key)){
				core.debug("recoveryQuat key:" + key + " not found.");
			}
			if(!progress.ContainsKey(type)){
				core.debug("progress key:" + type + " not found.");
			}
		}

		public float AdjustInterval(bool bForce){
			float stageCoefficient = (bForce) ? 5.0f * 0.1f : (float)core.getStage() * 0.1f;
			return (1 / (1 + stageCoefficient));
		}
		
		public float AdjustStrength(bool bForce){
			float stageCoefficient = (bForce) ? 5.0f * 0.1f : (float)core.getStage() * 0.2f;
			return 1 + stageCoefficient;
		}
		
		public void manage(float time){
			try{
				if(time == lastTime){
					time += 0.001f;
				}else if(time < lastTime){
					time = lastTime + 0.001f;
				}
				lastTime = time;

				core.guess.CheckTrunk();
				int stage = core.getStage();
				foreach(var key in keyCommand){
					if(isActive[key]){
						if(isForce[key] || stage > 0){
							calcProgress(time, key);
							changeJoint(time, keyHead, key);
						}else{
							isActive[key] = false;
						}
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void RandomHeadMotionForBigWave(float ratioMax = 1.0f, bool bForce = false){
			if(headMotionOnGlobal == false){
				return;
			}

			float threshold0 = 0.00f;//dummy
			
			float threshold1 = threshold0 + 0.05f;//glanceForce
			float threshold2 = threshold1 + 0.05f;//faceSidewaysForce
			float threshold3 = threshold2 + 0.03f;//faceSidewaysAltForce
			float threshold4 = threshold3 + 0.01f;//faceSidewaysMaxForce
			float threshold5 = threshold4 + 0.02f;//bendHerselfBackStrongForce
			if(core.orgasmCount == 0 && core.isFirstWave){
				threshold1 = threshold0 + 0.05f;//glanceForce
				threshold2 = threshold1 + 0.30f;//faceSidewaysForce
				threshold3 = threshold2 + 0.05f;//faceSidewaysAltForce
				threshold4 = threshold3 + 0.03f;//faceSidewaysMaxForce
				threshold5 = threshold4 + 0.30f;//bendHerselfBackStrongForce
			}else if(core.orgasmCount < 3){
				//nop
			}else if(core.orgasmCount >= 3){
				threshold1 = threshold0 + 0.10f;//glanceForce
				threshold2 = threshold1 + 0.10f;//faceSidewaysForce
				threshold3 = threshold2 + 0.05f;//faceSidewaysAltForce
				threshold4 = threshold3 + 0.03f;//faceSidewaysMaxForce
				threshold5 = threshold4 + 0.05f;//bendHerselfBackStrongForce				
			}else if(core.orgasmCount > 5){
				threshold1 = threshold0 + 0.03f;//glanceForce
				threshold2 = threshold1 + 0.03f;//faceSidewaysForce
				threshold3 = threshold2 + 0.10f;//faceSidewaysAltForce
				threshold4 = threshold3 + 0.10f;//faceSidewaysMaxForce
				threshold5 = threshold4 + 0.05f;//bendHerselfBackStrongForce				
			}else if(core.orgasmCount > 10){
				threshold1 = threshold0 + 0.03f;//glanceForce
				threshold2 = threshold1 + 0.03f;//faceSidewaysForce
				threshold3 = threshold2 + 0.10f;//faceSidewaysAltForce
				threshold4 = threshold3 + 0.15f;//faceSidewaysMaxForce
				threshold5 = threshold4 + 0.15f;//bendHerselfBackStrongForce				
			}
			
			float ratio = UnityEngine.Random.Range(0.0f,ratioMax);
			if(ratio < threshold1){
				glanceForce(bForce);
			}else if(ratio < threshold2){
				faceSidewaysForce(bForce);
			}else if(ratio < threshold3){
				faceSidewaysAltForce(bForce);
			}else if(ratio < threshold4){
				faceSidewaysMaxForce(bForce);				
			}else if(ratio < threshold5){
				bendHerselfBackStrongForce(bForce);
			}
		}

		public void PickAction(string stage, bool bIgnoringFiringRatio = false, bool bForce = false){
			float rnd = UnityEngine.Random.Range(0f, 1f);
			float ratio = GetFireRatio(stage);
			bool flag = (bIgnoringFiringRatio || rnd < ratio) ? true : false;
			if(flag){
				string actionName = stageAndEventMap.PickAction(stage, true);
				updateMotion(actionName, bForce);
			}
		}
		
		public void PickAction(bool bIgnoringFiringRatio = false, bool bForce = false){
			PickAction(core.getStage(), core.orgasmCount, bIgnoringFiringRatio);
		}
		
		public void PickAction(int stage, int count, bool bIgnoringFiringRatio = false, bool bForce = false){
			string name = ResolveStageName(stage, count);
			PickAction(name, bIgnoringFiringRatio);
		}
		
		public void RandomHeadMotionForSmallWave(float ratioMax = 1.0f, bool bForce = false){
			if(headMotionOnGlobal == false){
				return;
			}
			float ratio = UnityEngine.Random.Range(0.0f,ratioMax);
			if(ratio < 0.05f){
				RandomHeadMotionForBigWave(ratioMax:0.50f);
			}else if(ratio < 0.3f){
				bendHerselfBackForce(bForce);
			}
		}
		
		public void bendHerselfBack(bool bForce = false){
			float ratio = UnityEngine.Random.Range(0.0f,1.0f);
			if(ratio < bendBackwardThreshold){
				updateMotion("Turn Head Back 1", bForce);
			}
		}

		public void FaceForwardForce(bool bForce = false){
			updateMotion("Face Forward", bForce);
		}
		
		public void FaceForwardSlowlyForce(bool bForce = false){
			updateMotion("faceForwardSlowly", bForce);
		}
		
		public void bendHerselfBackForce(bool bForce = false){
			updateMotion("Turn Head Back 1", bForce);
		}
		
		public void bendHerselfBackStrongForce(bool bForce = false){
			updateMotion("Turn Head Back 2", bForce);
		}
		
		public void faceSidewaysForce(bool bForce = false, float strength = 0f){
			updateMotion("Face Sideways 1", bForce, strength);
		}
		
		public void faceSidewaysAltForce(bool bForce = false){
			updateMotion("Face Sideways 2", bForce);
		}
		
		public void faceSidewaysMaxForce(bool bForce = false){
			updateMotion("Face Sideways 3", bForce);
		}
		
		public void glanceForce(bool bForce = false){
			updateMotion("Glance", bForce);
		}
		
		public void OnDestroy(){
			RemoveBackup();
			RemoveTiming();
		}
		
		public void RemoveBackup(){
			foreach(var kvp in backup){
				if(backup[kvp.Key] != null){
					backup[kvp.Key] = null;
				}
			}
			backup = null;
		}
		
		public string ResolveStageName(int stage, int count){
			string name = (stage == 5) ? "Stage=" : "Stage<";
			name += (stage < 3) ? "3" : "5";
			if(stage == 5){
				int threshold = 99;
				if(count < 1){ threshold = 1; }
				else if(count < 3){ threshold = 3; }
				else if(count < 6){ threshold = 6; }
				else if(count < 9){ threshold = 9; }
				name += ":OrgasmCount<" + ((threshold == 99) ? threshold.ToString() : ("0" + threshold.ToString()));
			}else{
				name += ":OrgasmCount<99";
			}
			return name;
		}

		public void RemoveTiming(){
			foreach(var kvp in timing){
				if(timing[kvp.Key] != null){
					timing[kvp.Key] = null;
				}
			}
			timing = null;			
		}
		#endregion
	}
}