#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class MorphBlender{
		protected string[] names;
		public DAZMorph[] morphs;
		protected float[] limits;
		protected MorphManager mmE;
		protected MorphManager mmO;
		
		public MorphBlender(MorphManager mmE, MorphManager mmO){
			names = new string[3];
			morphs = new DAZMorph[3];
			limits = new float[3];
			this.mmE = mmE;
			this.mmO = mmO;
			for(int i = 0; i < 3; i++){
				names[i] = "";
			}
		}
		
		public void addExpression(string name, DAZMorph morph, float limit){
			if(names[2] != ""){
				if(morphs[2] != null){
					morphs[2].SetValue(0f);
				}
			}
			for(int i = 2; i > 0; i--){
				if(names[i-1] == ""){
					continue;
				}
				
				if(morphs[i-1] != null){
					morphs[i] = morphs[i-1];
				}
				names[i]  = names[i-1];
				limits[i] = limits[i-1];
			}
			
			if(morph != null){
				morphs[0] = morph;
			}
			names[0]  = name;
			limits[0] = limit;
		}
		
		public void setRandomMorph(int stage){
			string name;
			DAZMorph morph;
			float limit;
			MorphManager mm = (stage < 5) ? mmE : mmO;

			name  = mm.getRandomMorphName();
			morph = mm.ms[name].morph;
			limit = mm.ms[name].limit;
			addExpression(name, morph, limit);
		}
		
		public float getLimit(){
			return limits[0];
		}
		
		public void changeExpression(float strength, float ratio){
			float ratioAlt = 1 - ratio;
			if(names[1] == "" || names[0] == names[1]){
				if(morphs[0] != null){
					morphs[0].SetValue(strength);
				}
			}else{
				if(morphs[0] != null){
					morphs[0].SetValue(strength * ratio);
					morphs[1].SetValue(strength * ratioAlt);					
				}
			}
		}
		
		public void stopExpression(){
			mmE.stopExpression();
			mmO.stopExpression();			
		}

		public void debug(string str){
			SuperController.LogMessage("debug: " + str);
		}
		
		public void OnDestroy(){
			mmE.OnDestroy();
			mmE = null;
			
			mmO.OnDestroy();
			mmO = null;
		}
	}	
	
	public class MorphManager{
		protected Atom atom;
		protected List<string> enabledMorphs;
		protected string[] allMorphs;
		protected GenerateDAZMorphsControlUI morphControl;
		public Dictionary<string, MorphSetting> ms;
		protected string morphListTag;
		protected string morphSettingTag;
		protected JSONNode jsnroot;
		
		public MorphManager(){
			jsnroot = new JSONNode();
			ms = new Dictionary<string, MorphSetting>();
		}
		
		public void initialize(Atom atom, JSONNode jsnroot, string morphListTag, string morphSettingTag){
			try{
				this.atom = atom;
				JSONStorable geometry = atom.GetStorableByID("geometry");
				DAZCharacterSelector character = geometry as DAZCharacterSelector;
				morphControl = character.morphsControlUI;
			
				this.morphListTag = morphListTag;
				this.morphSettingTag = morphSettingTag;
				this.jsnroot = jsnroot;
				
				JSONNode jsn = this.jsnroot[this.morphListTag];
				for(int i = 0; i < jsn.Count; i++){
					string name = jsn[i];
					DAZMorph morph = morphControl.GetMorphByDisplayName(name);
					if(name == "" || ms.ContainsKey(name)) continue;
					ms.Add(name, new MorphSetting(name, morph));
				}
				
				repackEnabledMorphs();
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
			}
		}
		
		public void overrideSettingsAll(){
			try{
				JSONNode jsn = jsnroot[morphSettingTag];
				if(jsn == null){
					return;
				}
				int num = jsn.Count;
				for(int i = 0; i < num; i++){
					string name = jsn[i]["Name"];
					ms[name].morph   = morphControl.GetMorphByDisplayName(name);
					ms[name].enabled = jsn[i]["Enabled"].AsBool;
					ms[name].limit   = jsn[i]["Limit"].AsFloat;
				}
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
			}
		}
		
		public void overrideSetting(string name, bool enabled, float limit){
			try{
				ms[name].morph   = morphControl.GetMorphByDisplayName(name);
				ms[name].enabled = enabled;
				ms[name].limit   = limit;
				
				repackEnabledMorphs();
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
			}
		}

		public void rewriteJSON(){
			try{
				foreach(var kvp in ms){
					jsnroot[morphSettingTag][kvp.Key]["Name"] = ms[kvp.Key].name;
					jsnroot[morphSettingTag][kvp.Key]["Enabled"] = ms[kvp.Key].enabled.ToString();
					jsnroot[morphSettingTag][kvp.Key]["Limit"] = ms[kvp.Key].limit.ToString();
				}
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
			}
		}
				
		public void stopExpression(){
			try{
				for(int i = 0; i < enabledMorphs.Count; i++){
					if(ms[enabledMorphs[i]].morph != null){
						ms[enabledMorphs[i]].morph.SetValue(0.000001f);
					}
				}
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
			}
		}
				
		public void repackEnabledMorphs(){
			try{
				enabledMorphs = new List<string>();
				foreach(var kvp in ms){
					if(ms[kvp.Key].enabled){
						enabledMorphs.Add(kvp.Key);
					}
				}
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
			}
		}
		
		public string getRandomMorphName(){
			try{
				int num = UnityEngine.Random.Range(0, enabledMorphs.Count - 1);
				return ms[enabledMorphs[num]].name;
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
				return null;
			}
		}
		
		public void OnDestroy(){
			RemoveAllMorphSettings();
		}
		
		public void RemoveAllMorphSettings(){
			foreach(var kvp in ms){
				if(ms[kvp.Key] != null){
					ms[kvp.Key] = null;
				}
			}
			ms = null;
		}
		
		public void debug(string str){
			SuperController.LogMessage("debug: " + str);
		}
	}
	
	public class MorphSetting{
		public string name;
		public bool enabled;
		public float limit;
		public DAZMorph morph;
		
		public MorphSetting(string name, DAZMorph morph){
			this.name = name;
			this.morph = morph;
			limit = 1.0f;
			enabled = true;
		}
	}
}