using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using Object = UnityEngine.Object;
using System.Text.RegularExpressions;

namespace Cotyounoyume
{
    public class AddonTakeControlItem : MonoBehaviour 
    {
		public Atom atom;
		public EBATCore core;
		private float lastTime;
		public bool autoStopOnGlobal = true;
		
		protected List<string> potentialConflictsTable;
		
		protected Dictionary<string, bool> statusList;
		
		#region takeControlOfExpression
		ReceiverBackup receiverBackup;
		VariableTriggerBackup variableTriggerBackup;
		CollisionTriggerBackup collisionTriggerBackup;
		#endregion
		
		#region addon common
		public AddonTakeControlItem(EBATCore core){
			atom = core.atom;
			this.core = core;
			statusList = new Dictionary<string, bool>();
			potentialConflictsTable = new List<string>();
		}
		
		public void loadSettings(){
			try{
				autoStopOnGlobal = core.settings["General"]["AutoStop OtherPlugins"].AsBool;
				potentialConflictsTable.Clear();
				foreach(var name in core.settingsRoot["Common"]["Plugins that might Confilict"].AsArray){
					potentialConflictsTable.Add(name.ToString().Trim('"'));
				}
				if(core.settingsRoot["Common"].AsObject.HasKey("PTMC On/Off")){
					foreach(var name in potentialConflictsTable){
						if(!core.settingsRoot["Common"]["PTMC On/Off"].AsObject.HasKey(name)){
							core.settingsRoot["Common"]["PTMC On/Off"][name] = "true";
						}
					}
				}else{
					foreach(var name in potentialConflictsTable){
						core.settingsRoot["Common"]["PTMC On/Off"][name] = "true";
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void applySettings(){
		}

		public void manage(float time){
			/*
			if(time == lastTime){
				time += 0.001f;
			}else if(time < lastTime){
				time = lastTime + 0.001f;
			}
			lastTime = time;
			*/
		}	
		
		public void OnDestroy(){
			ReturnControlOfPlugins();
		}
		
		public void OnOrgasmStart()
		{
		}
		
		public void OnOrgasmStop()
		{
		}
		#endregion
		
		#region TakeControl
		public void test(){
			List<string> ids = atom.GetStorableIDs();
			foreach(var id in ids){
				if(id.StartsWith("plugin#")){
					core.debug("plugin=> " + id);
					bool bHit = false;
					foreach(var item in potentialConflictsTable){
						if(id.Contains(item)){
							bHit = true;
						}
					}
					if(bHit){
						core.debug("plugin:" + id + " is plugin with potential Conflicts.");
					}else{
						core.debug("plugin:" + id + " is NOT plugin with potential Conflicts.");
					}
					
					if(bHit){
						JSONStorable jss = atom.GetStorableByID(id);
						jss.SetBoolParamValue("enabled", false);
					}
					/*
					JSONStorable jss = atom.GetStorableByID(id);
					List<string> names = jss.GetBoolParamNames();
					foreach(var name in names){
						core.debug(id + ": " + name);
					}
					*/
				}
			}
		}
		
		public List<string> GetPluginList(string mode = "unlisted"){
			List<string> result = new List<string>();
			List<string> ids = atom.GetStorableIDs();
			foreach(var id in ids){
				if(id.StartsWith("plugin#")){
					string pluginName = NormalizePluginName(id);
					if(mode == "unlisted" && !isPluginListed(pluginName)){
						result.Add(NormalizePluginName(id));
					}else if(mode == "all"){
						result.Add(NormalizePluginName(id));
					}
				}
			}
			return result;
		}
		
		public string NormalizePluginName(string orig){
			string[] pattern = {"_"};
			string[] words = orig.Split(pattern, StringSplitOptions.None);
			string pluginPrefix = words[0];
			string pluginName = orig.Remove(0, pluginPrefix.Length+1);
			return pluginName;
		}
		
		public void AutoStopPlugins(){
			if(core.isHot){
				TakeControlOfPlugins();
			}else{
				ReturnControlOfPlugins();
			}
		}
		
		public void TakeControlOfPlugins(){
			List<string> ids = atom.GetStorableIDs();
			foreach(var id in ids){
				if(id.StartsWith("plugin#")){
					string pluginName = NormalizePluginName(id);
					if(isPluginWithPotentialConflits(pluginName)){
						TemporarilyDeactivatePlugin(id);
					}
				}
			}
		}
		
		public bool isPluginListed(string id){
			if(core.settingsRoot["Common"]["PTMC On/Off"].AsObject.HasKey(id)){
				return true;
			}else{
				return false;
			}
		}
		
		public bool isPluginWithPotentialConflits(string id){
			if(isPluginListed(id) && core.settingsRoot["Common"]["PTMC On/Off"][id].AsBool){
				return true;
			}else{
				return false;
			}
		}

		public void TemporarilyDeactivatePlugin(string id){
			JSONStorable jss = atom.GetStorableByID(id);
			bool currentStatus = jss.GetBoolParamValue("enabled");
			if(statusList.ContainsKey(id)){
//				statusList[id] = currentStatus;
			}else{
				statusList.Add(id, currentStatus);				
			}
			jss.SetBoolParamValue("enabled", false);
		}
		
		public void ReturnControlOfPlugins(){
			foreach(KeyValuePair<string, bool> kvp in statusList){
				ReavtivatePlugin(kvp.Key);
			}
			statusList.Clear();
		}
		
		public void ReavtivatePlugin(string id){
			JSONStorable jss = atom.GetStorableByID(id);
			jss.SetBoolParamValue("enabled", statusList[id]);
		}
		#endregion
		
		#region TakeControl
		public void TakeControlOfExpression(){
			Dictionary<VariableTrigger, JSONClass> res1 = Utility.CollectVariableTrigger();
			variableTriggerBackup = new VariableTriggerBackup(res1, atom);
			variableTriggerBackup.ClearTargetMorphs();
			
			Dictionary<CollisionTrigger, JSONClass> res2 = Utility.CollectCollisionTrigger();
			collisionTriggerBackup = new CollisionTriggerBackup(res2, atom);
			collisionTriggerBackup.ClearTargetMorphs();
			
//			Dictionary<Trigger, JSONClass> res = Utility.CollectAnimationTrigger();
		}
		
		public void ReturnControlOfExpression(){
			variableTriggerBackup.Recover();
			collisionTriggerBackup.Recover();
		}				
		#endregion
	}
}