#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class ClothesStateChanger{
		List<ClothStateChanger> clothStateChangerArray = new List<ClothStateChanger>();
		DAZClothingItem test;
		Atom atom;
		DAZSkinWrap skinwrap1;
		List<DAZClothingItem> myClothes;
		
		public ClothesStateChanger(Atom atom){
			try{
				JSONStorable geometryStorable = atom.GetStorableByID("geometry");
				DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
				
				this.atom = atom;
				myClothes = new List<DAZClothingItem>();
				
				List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
				for (int i = 0; i < clothes.Count; i++){
					DAZClothingItem c = clothes[i];
					if(c.active && c.creatorName != "cotyounoyume"){
						clothStateChangerArray.Add(new ClothStateChanger(c));
						myClothes.Add(c);
					}
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void ChangeClothes(){
			myClothes.Clear();
			clothStateChangerArray.Clear();
			
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			
			myClothes = new List<DAZClothingItem>();
			clothStateChangerArray = new List<ClothStateChanger>();
			
			List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
			for (int i = 0; i < clothes.Count; i++){
				DAZClothingItem c = clothes[i];
				if(c.active && c.creatorName != "cotyounoyume"){
					clothStateChangerArray.Add(new ClothStateChanger(c));
					myClothes.Add(c);
				}
			}
		}
		
		public void RemoveClothes(){
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			foreach(var cloth in myClothes){
				if(cloth.creatorName != "cotyounoyume"){
					geometryForClothes.SetActiveClothingItem(cloth, active: false);
				}
			}
		}
		
		public void RemoveResetClothes(){
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			foreach(var cloth in myClothes){
				geometryForClothes.SetActiveClothingItem(cloth, active: true);
			}
		}
		
		public void CheckClothAttribute(){
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			foreach(var cloth in myClothes){
				Utility.debug("clothname:" + cloth.displayName);
				Utility.debug("  tags:" + cloth.tags);
				Utility.debug("  packageUid:" + cloth.packageUid);
				Utility.debug("  type:" + cloth.type);
				Utility.debug("  hasCustomizationUI:" + cloth.hasCustomizationUI);
				DAZClothingItemControl[] componentsInChildren = cloth.GetComponentsInChildren<DAZClothingItemControl>(includeInactive: true);
				if(componentsInChildren == null){
					Utility.debug("  componentsInChildren: null");
				}else{
					Utility.debug("  componentsInChildren: not null, length:" + componentsInChildren.Length);
					DAZClothingItemControl dcic = componentsInChildren[0];
				}
			}
		}
		
		public void setTestCloth(string displayName){
			DAZClothingItem test = Utility.getClothByDisplayName(atom, displayName);
			Utility.debug("test");
			if(test == null){
				Utility.debug("test is null");												
			}else{
				Utility.debug("test is not null");																
			}
			Utility.debug(displayName + ": isDynamicRuntimeLoaded:" + test.isDynamicRuntimeLoaded);
			Utility.debug(displayName + ": isLegacy:" + test.isLegacy);
			
			
			ClothSimControl[] cscArray = test.GetComponentsInChildren<ClothSimControl>();
			if(cscArray == null){
				Utility.debug("cscArray is null");				
				return;
			}else{
				Utility.debug("cscArray is not null");				
			}
			if(cscArray.Length == 0){
				Utility.debug("cscArray is null2");
				return;
			}
			Utility.debug("cscArray.Length:" + cscArray.Length);
			ClothSimControl csc1 = cscArray[0];
			csc1.clothSettings.BreakThreshold = 0f;
			Utility.debug("debug");
			skinwrap1 = csc1.clothSettings.MeshProvider.PreCalcProvider as DAZSkinWrap;
			if(csc1 != null){
				Utility.debug("set skinwrap1");
			}
		}
		
		public void Recover(){
			try{
				for(int i = 0; i < clothStateChangerArray.Count; i++){
					clothStateChangerArray[i].Recover();
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void Reset(){
			try{
				for(int i = 0; i < clothStateChangerArray.Count; i++){
					clothStateChangerArray[i].Reset();
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void Change(string type = "UndressStrong", float ratio = 1.0f){
			try{
				for(int i = 0; i < clothStateChangerArray.Count; i++){
					clothStateChangerArray[i].Change(type, ratio);
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void UndressStop(){
			try{
				for(int i = 0; i < clothStateChangerArray.Count; i++){
					clothStateChangerArray[i].UndressStop();
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void OnDestroy(){
			for(int i = 0; i < clothStateChangerArray.Count; i++){
				clothStateChangerArray[i].OnDestroy();
				clothStateChangerArray[i] = null;
			}
			clothStateChangerArray = null;
		}
	}
	
	public class ClothStateChanger
	{
		public List<ClothSimStateChanger> cssc = new List<ClothSimStateChanger>();
		private DAZClothingItem cloth;
		
		public ClothStateChanger(DAZClothingItem cloth){
			this.cloth = cloth;
			ClothSimControl[] cscArray = this.cloth.GetComponentsInChildren<ClothSimControl>();
			foreach (ClothSimControl csc in cscArray){
				cssc.Add(new ClothSimStateChanger(csc));
			}
		}
		
		public void Recover(){
			for(int i = 0; i < cssc.Count; i++){
				cssc[i].Recover();
			}
		}
		
		public void Reset(){
			for(int i = 0; i < cssc.Count; i++){
				cssc[i].Reset();
			}
		}
		
		public void Change(string type, float ratio){
			for(int i = 0; i < cssc.Count; i++){
				cssc[i].Change(type, ratio);
			}
		}
		
		public void UndressStop(){
			for(int i = 0; i < cssc.Count; i++){
				cssc[i].UndressStop();
			}
		}
		
		public void OnDestroy(){
			for(int i = 0; i < cssc.Count; i++){
				cssc[i].OnDestroy();
				cssc[i] = null;
			}
			cssc = null;
		}
	}
	
	public class ClothSimStateChanger
	{
		public ClothSimControl csc;
		public bool    backupBreakEnabled;
		public float   backupBreakThreshold;
		public float   backupFriction;
		public float   backupWeight;
		public float   backupGravityMultiplier;
		public float   backupStiffness;
		public bool    backupSimEnabled;
				
		public ClothSimStateChanger(ClothSimControl csc){
			this.csc = csc;
			backupBreakEnabled      = this.csc.clothSettings.BreakEnabled;
			backupBreakThreshold    = this.csc.clothSettings.BreakThreshold;
			backupFriction          = this.csc.clothSettings.Friction;
			backupWeight            = this.csc.clothSettings.Weight;
			backupGravityMultiplier = this.csc.clothSettings.GravityMultiplier;
			backupStiffness         = this.csc.clothSettings.Stiffness;
			backupSimEnabled        = this.csc.simEnabledJSON.val;
		}	

		public void Recover(){
			csc.clothSettings.BreakEnabled      = backupBreakEnabled;
			csc.clothSettings.BreakThreshold    = backupBreakThreshold;
			csc.clothSettings.Friction          = backupFriction;
			csc.clothSettings.Weight            = backupWeight;		
			csc.clothSettings.GravityMultiplier = backupGravityMultiplier;
			csc.clothSettings.Stiffness         = backupStiffness;
			csc.SetSimEnabled(backupSimEnabled);
		}
		
		public void Reset(){
			csc.Reset();
		}

		public void UndressStop(){
			csc.clothSettings.Friction          = 1.0f;
			csc.clothSettings.Weight            = 0.5f;
			csc.clothSettings.GravityMultiplier = 1.0f;
			csc.clothSettings.Stiffness = 1f;
			csc.clothSettings.BreakEnabled   = false;
			csc.clothSettings.BreakThreshold = 1f;
			
/*
			DAZSkinWrap skinWrap = csc.clothSettings.MeshProvider.PreCalcProvider as DAZSkinWrap;
			if(skinWrap != null){
				
				DAZClothingItem test = Utility.getClothByDisplayName(atom, displayName);

				ClothSimControl[] cscArray = test.GetComponentsInChildren<ClothSimControl>();
				Utility.debug("cscArray is null");				
				if(cscArray == null){
					Utility.debug("cscArray is null");				
					return;
				}
				ClothSimControl csc1 = cscArray[0];
				DAZSkinWrap skinwrap1 = csc1.clothSettings.MeshProvider.PreCalcProvider as DAZSkinWrap;
				if(csc1 != null){
					Utility.debug("set skinwrap1");
				}				
				
				csc.clothSettings.MeshProvider.PreCalcProvider = skinwrap1;
				/*
				Utility.debug("csc.skinWrap.IsWrapping: " + skinWrap.IsWrapping);
				Utility.debug("csc.skinWrap.WrapStatus: " + skinWrap.WrapStatus);
				Utility.debug("csc.skinWrap.simTextures.Length: " + skinWrap.simTextures.Length);
				Utility.debug("csc.skinWrap.numMaterials: " + skinWrap.numMaterials);
				Utility.debug("csc.skinWrap.wrapName: " + skinWrap.wrapName);
				if(skinWrap.numMaterials > 0){
					for(int i = 0; i < skinWrap.numMaterials; i++){
						Utility.debug("materialNames: " + i + ": " + skinWrap.materialNames[i]);
					}
				}
				if(skinWrap.simTextures.Length > 0){
					for(int i = 0; i < skinWrap.simTextures.Length; i++){
						Utility.debug("before");
						skinWrap.simTextures[i] = Texture2D.blackTexture;
						Utility.debug("after");
					}
				}
				skinWrap.wrapStore.StoreToBinaryFile("Custom/Scripts/mytest.dat");
				skinWrap.wrapStore.LoadFromBinaryFile("Custom/Scripts/mytest.dat");
			}else{
				Utility.debug("can't get skinWrap.");
				if(csc.clothSettings != null){
					Utility.debug("csc has csc.clothSettings");
				}
				if(csc.clothSettings.MeshProvider != null){
					Utility.debug("csc has csc.clothSettings.MeshProvider");
				}
			}
			/*
			for(int i = 0; i < csc.skinWrap.simTextures.Length; i++){
				csc.skinWrap.
			}
			csc.skinWrap.simTextures
			*/
		}
		
		public void Change(string type, float ratio){
			if(type == "Undress"){
				csc.SetSimEnabled(true);
				csc.clothSettings.BreakEnabled   = true;
				csc.clothSettings.BreakThreshold = 0f;					
				Undress(ratio);
			}else if(type == "Break"){
				csc.SetSimEnabled(true);
				csc.clothSettings.BreakEnabled   = true;
				csc.clothSettings.BreakThreshold = 0.1f;
				csc.SetSimEnabled(true);
				csc.SyncAllowDetatch(true);
			}else if(type == "Break2"){
				csc.SetSimEnabled(true);
				csc.clothSettings.BreakEnabled   = true;
				csc.clothSettings.BreakThreshold = 0.0f;
				csc.SetSimEnabled(true);
				csc.SyncAllowDetatch(true);
			}else if(type == "Reverse"){
				Reverse(ratio);
				csc.clothSettings.Stiffness = 1f;
			}else if(type == "UndressAndReverse"){
				csc.SetSimEnabled(true);
				csc.clothSettings.BreakEnabled   = true;
				csc.clothSettings.BreakThreshold = 0f;			
				csc.clothSettings.Stiffness = 1f;
				Reverse(ratio);
			}else{
				Utility.debug("In ClothSimControl.Change: Unknown Command: " + type);
			}
		}
		
		public void Undress(float ratio){
			csc.clothSettings.Friction          = Mathf.Lerp(1.0f, 0.1f, ratio);
			csc.clothSettings.Weight            = Mathf.Lerp(0.5f, 2.0f, ratio);
			csc.clothSettings.GravityMultiplier = Mathf.Lerp(0.5f, 2.0f, ratio);			
		}
		
		public void Reverse(float ratio){
			csc.clothSettings.Friction          = Mathf.Lerp(1.0f,  0.1f, ratio);
			csc.clothSettings.Weight            = Mathf.Lerp(0.5f,  2.0f, ratio);
			csc.clothSettings.GravityMultiplier = Mathf.Lerp(1.0f, -2.0f, ratio);
		}
				
		public void OnDestroy(){
			csc = null;
		}
	}

}